! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
module gofio_mod

  use dal
  use errorHandling
  use gofutils_mod
  implicit none

  public 

contains

  subroutine check_open_set(in_F,out_S,mode,status,warndie)
    !
    ! sub to open FITS file properly
    !
    type(DataSetT), intent(out) :: out_S    
    character(len=*), intent(in) :: in_F
    character(len=1) :: cmode
    character(len=1024) :: str1, str2, str3, str4 ! fix this with in_F
    integer, intent(in) :: mode ! 1=read, 2=write, 3=create, 4=modify
    integer, intent(inout) :: status
    logical :: exists = .FALSE.
    ! warndie = .TRUE. = fatal 
    ! warndie = .FALSE. = dependent on type. CAN continue if file not critical
    logical, intent(inout) :: warndie
    !
    write(cmode,'(I1)')mode
    exists = setExists(trim(in_F))
!    write(errstr,*)"check_open_set: Does ",trim(in_F)," exist? exist = ",exists
!    call message(AppMsg,Verbose) 
    write(str1,*) "Critical input file: ",trim(in_F)," not found/does not exist. Exiting..."
    write(str2,*) "Non-Critical input file: ",trim(in_F)," not found/does not exist, but can continue..."
    write(str3,*) "Opening file: ",trim(in_F)," for READ(1)/WRITE(2)/CREATE(3)/MODIFY(4):",cmode
    write(str4,*) "Input file: ",trim(in_F)," exists, but SAS_CLOBBER = NO. Exiting..."
    !
    select case (mode)
    case(1) ! read only
       if (exists) then 
        !  write(errstr,*)trim(str3)
        !  call message(AppMsg,Noisy) ! why doesn't this work???!!!
          out_S = dataSet(trim(in_F), READ)
          status = 1
       else
          if (warndie) then
             call error("NoInfile",trim(str1))          
          else
             call warning("NoInfile",trim(str2))
          endif
       end if
    case(2) ! write
       if (exists) then
          if (clobber()) then ! if noclobber can one still WRITE??
         !    write(errstr,*)str3
         !    call message(AppMsg,Noisy)
             out_S = dataSet(trim(in_F), WRITE)
             status = 1
          else
             call error("noClobber",trim(str4))
          endif
       else 
          if (warndie) then
             call error("NoInfile",trim(str1))          
          else
             call warning("NoInfile",trim(str2))
          endif
       end if
    case(3) ! create
       if (exists) then
          if (clobber()) then 
          !   write(errstr,*)str3
          !   call message(AppMsg,Noisy)
             out_S = dataSet(trim(in_F), CREATE) ! dal error if can't open
             status = 1
          else
             call error("noClobber",trim(str4))
          endif
       else 
          out_S = dataSet(trim(in_F), CREATE) ! just open it
          status = 1
       end if
    case(4) ! modify (why different than write???
       if (exists) then
          if (clobber()) then ! if noclobber can one still WRITE??
           !  write(errstr,*)str3
           !  call message(AppMsg,Noisy)
             out_S = dataSet(trim(in_F), MODIFY)
             status = 1
          else
             call error("noClobber",trim(str4))
          endif
       else 
          if (warndie) then
             call error("NoInfile",trim(str1))          
          else
             call warning("NoInfile",trim(str2))
          endif
       end if
    case DEFAULT
       write(errstr,*) "File open mode",mode,&
            " not supported (must be 1-4)" ! should never happen!!!
       call error("badMode",trim(errstr))
    end select
    warndie = .TRUE.
    
    return

  end subroutine check_open_set
  !
  !
  !
  subroutine coord_keys (in_S, ra_pnt, dec_pnt, pa_pnt, crval, cdelt, &
       cunit, ctype, crpix, crota, radecsys, warndie)
    !
    ! extract just coordinate system/WCS keywords from image files
    !
    type(DataSetT), intent(in) :: in_S
    type(ArrayT) :: in_A
    character(len=*), intent(out) :: cunit(2), ctype(2), radecsys
    logical :: warndie
    real(kind=double), intent(out) :: crval(2), cdelt(2), crpix(2), &
         crota(2), ra_pnt, dec_pnt, pa_pnt

    in_A = array(in_S, 0, READ) ! set already opened in check_open_set

    if (hasAttribute(in_A,"RA_PNT")) then
       ra_pnt = real64Attribute(in_A,"RA_PNT")
    elseif (hasAttribute(in_S,"RA_PNT")) then
       ra_pnt = real64Attribute(in_S,"RA_PNT")
    else
       if (warndie) then
          call error ("imgNoRA_PNT",&
               "coord_keys: File does not have RA_PNT attribute")
       else
          call warning ("imgNoRA_PNT",&
               "coord_keys: File does not have RA_PNT attribute")
       end if
    end if
    if (hasAttribute(in_A,"DEC_PNT")) then
       dec_pnt = real64Attribute(in_A,"DEC_PNT")
    elseif (hasAttribute(in_S,"DEC_PNT")) then
       dec_pnt = real64Attribute(in_S,"DEC_PNT")
    else
       if (warndie) then
          call error ("imgNoDEC_PNT",&
               "coord_keys: File does not have DEC_PNT attribute")
       else
          call warning ("imgNoDEC_PNT",&
               "coord_keys: File does not have DEC_PNT attribute")
       end if
    end if
    if (hasAttribute(in_A,"PA_PNT")) then
       pa_pnt = real64Attribute(in_A,"PA_PNT")
    elseif (hasAttribute(in_S,"PA_PNT")) then
       pa_pnt = real64Attribute(in_S,"PA_PNT")
    else
       if (warndie) then
          call error ("imgNoPA_PNT",&
               "coord_keys: File does not have PA_PNT attribute")
       else
          call warning ("imgNoPA_PNT",&
               "coord_keys: File does not have PA_PNT attribute")
       end if
    end if

    if (hasAttribute(in_A,"CRVAL1")) then
       crval(1) = real64Attribute(in_A,"CRVAL1")
    elseif (hasAttribute(in_S,"CRVAL1")) then
       crval(1) = real64Attribute(in_S,"CRVAL1")
    else
       if (warndie) then
          call error ("imgNoCRVAL",&
               "coord_keys: File does not have CRVAL1 attribute")
       else
          call warning ("imgNoCRVAL",&
               "coord_keys: File does not have CRVAL1 attribute")
       end if
    end if

    if (hasAttribute(in_A,"CRVAL2")) then
       crval(2) = real64Attribute(in_A,"CRVAL2")
    elseif (hasAttribute(in_S,"CRVAL2")) then
       crval(2) = real64Attribute(in_S,"CRVAL2")
    else
       if (warndie) then
          call error ("imgNoCRVAL",&
               "coord_keys: File does not have CRVAL2 attribute")
       else
          call warning ("imgNoCRVAL",&
               "coord_keys: File does not have CRVAL2 attribute")
       end if
    end if

    if (hasAttribute(in_A,"CDELT1")) then
       cdelt(1) = real64Attribute(in_A,"CDELT1")
    elseif (hasAttribute(in_S,"CDELT1")) then
       cdelt(1) = real64Attribute(in_S,"CDELT1")
    else
       if (warndie) then
          call error ("imgNoCDELT",&
               "coord_keys: File does not have CDELT1 attribute")
       else
          call warning ("imgNoCDELT",&
               "coord_keys: File does not have CDELT1 attribute")
       end if
    end if

    if (hasAttribute(in_A,"CDELT2")) then
       cdelt(2) = real64Attribute(in_A,"CDELT2")
    elseif (hasAttribute(in_S,"CDELT2")) then
       cdelt(2) = real64Attribute(in_S,"CDELT2")
    else
       if (warndie) then
          call error ("imgNoCDELT",&
               "coord_keys: File does not have CDELT2 attribute")
       else
          call warning ("imgNoCDELT",&
               "coord_keys: File does not have CDELT2 attribute")
       end if
    end if

    if (hasAttribute(in_A,"CUNIT1")) then
       cunit(1) = stringAttribute(in_A,"CUNIT1") 
    elseif (hasAttribute(in_S,"CUNIT1")) then
       cunit(1) = stringAttribute(in_A,"CUNIT1")
    else
       if (warndie) then
          call error ("imgNoCUNIT",&
               "coord_keys: File does not have CUNIT1 attribute")
       else
          call warning ("imgNoCUNIT",&
               "coord_keys: File does not have CUNIT1 attribute")
       end if
    end if

    if (hasAttribute(in_A,"CUNIT2")) then
       cunit(2) = stringAttribute(in_A,"CUNIT2") 
    elseif (hasAttribute(in_S,"CUNIT2")) then
       cunit(2) = stringAttribute(in_S,"CUNIT2") 
    else
       if (warndie) then
          call error ("imgNoCUNIT",&
               "coord_keys: File does not have CUNIT2 attribute")
       else
          call warning ("imgNoCUNIT",&
               "coord_keys: File does not have CUNIT2 attribute")
       end if
    end if

    if (hasAttribute(in_A,"CTYPE1")) then
       ctype(1) = stringAttribute(in_A,"CTYPE1")
    elseif (hasAttribute(in_S,"CTYPE1")) then
       ctype(1) = stringAttribute(in_A,"CTYPE1")
    else
       if (warndie) then
          call error ("imgNoCTYPE",&
               "coord_keys: File does not have CTYPE1 attribute")
       else
          call warning ("imgNoCTYPE",&
               "coord_keys: File does not have CTYPE1 attribute")
       end if
    end if

    if (hasAttribute(in_A,"CTYPE2")) then
       ctype(2) = stringAttribute(in_A,"CTYPE2")
    elseif (hasAttribute(in_S,"CTYPE2")) then
       ctype(2) = stringAttribute(in_A,"CTYPE2")
    else
       if (warndie) then
          call error ("imgNoCTYPE",&
               "coord_keys: File does not have CTYPE2 attribute")
       else
          call warning ("imgNoCTYPE",&
               "coord_keys: File does not have CTYPE2 attribute")
       end if
    end if

    if (hasAttribute(in_A,"CRPIX1")) then
       crpix(1) = real64Attribute(in_A,"CRPIX1")
    elseif (hasAttribute(in_S,"CRPIX1")) then
       crpix(1) = real64Attribute(in_S,"CRPIX1")
    else
       if (warndie) then
          call error ("imgNoCRPIX",&
               "coord_keys: File does not have CRPIX1 attribute")
       else
          call warning ("imgNoCRPIX",&
               "coord_keys: File does not have CRPIX1 attribute")
       end if
    end if

    if (hasAttribute(in_A,"CRPIX2")) then
       crpix(2) = real64Attribute(in_A,"CRPIX2")
    elseif (hasAttribute(in_S,"CRPIX2")) then
       crpix(2) = real64Attribute(in_S,"CRPIX2")
    else
       if (warndie) then
          call error ("imgNoCRPIX",&
               "coord_keys: File does not have CRPIX2 attribute")
       else
          call warning ("imgNoCRPIX",&
               "coord_keys: File does not have CRPIX2 attribute")
       end if
    end if

    write(errstr,*) "coord_keys: image CDELT(1,2) = ",cdelt(1),cdelt(2)
    call message(AppMsg,Noisy)

    if (hasAttribute(in_A,"CROTA1")) then
       crota(1) = real64Attribute(in_A,"CROTA1")
    elseif (hasAttribute(in_S,"CROTA1")) then
       crota(1) = real64Attribute(in_S,"CROTA1")
    else
       call warning ("imgNoCROTAn",&
            "coord_keys: File does not have CROTAn attribute")
       crota(1) = 0.
    end if
    if (hasAttribute(in_A,"CROTA2")) then
       crota(2) = real64Attribute(in_A,"CROTA2")
    elseif (hasAttribute(in_S,"CROTA2")) then
       crota(2) = real64Attribute(in_S,"CROTA2")
    else
       call warning ("imgNoCROTAn",&
            "coord_keys: File does not have CROTAn attribute")
       crota(2) = 0.
    end if

    if (hasAttribute(in_A,"RADECSYS")) then
       radecsys = stringAttribute(in_A,"RADECSYS")
    elseif (hasAttribute(in_S,"RADECSYS")) then
       radecsys = stringAttribute(in_S,"RADECSYS")
    else
       if (warndie) then
          call error ("imgNoRadecsys",&
               "coord_keys: File does not have RADECSYS attribute")
       else
          call warning ("imgNoRadecsys",&
               "coord_keys: File does not have RADECSYS attribute")
       end if
    end if

    return

  end subroutine coord_keys
  !
  !
  !
  subroutine more_coord_keys (in_S,xrefpix,yrefpix,xpixdelt,ypixdelt,warndie)
    !
    ! extract more coordinate system/WCS keywords from image files
    !
    type(DataSetT), intent(in) :: in_S
    type(ArrayT) :: in_A
    logical :: warndie
    real(kind=double), intent(out) :: xrefpix, yrefpix, xpixdelt, ypixdelt

    in_A = array(in_S, 0, READ) ! set already opened in check_open_set

    if (hasAttribute(in_A,"REFXCRPX")) then
       xrefpix = real64Attribute(in_A,"REFXCRPX")
    elseif (hasAttribute(in_S,"REFXCRPX")) then
       xrefpix = real64Attribute(in_S,"REFXCRPX")
    else
       if (warndie) then
          call error ("noREFXCRPX",&
               "more_coord_keys: File does not have REFXCRPX attribute")
       else
          call warning ("imgNoRA_PNT",&
               "more_coord_keys: File does not have REFXCRPX attribute")
       end if
    end if
    if (hasAttribute(in_A,"REFYCRPX")) then
       yrefpix = real64Attribute(in_A,"REFYCRPX")
    elseif (hasAttribute(in_S,"REFXCRPX")) then
       yrefpix = real64Attribute(in_S,"REFYCRPX")
    else
       if (warndie) then
          call error ("noREFYCRPX",&
               "more_coord_keys: File does not have REFYCRPX attribute")
       else
          call warning ("noREFYCRPX",&
               "more_coord_keys: File does not have REFYCRPX attribute")
       end if
    end if
    if (hasAttribute(in_A,"REFXCDLT")) then
       xpixdelt = real64Attribute(in_A,"REFXCDLT")
    elseif (hasAttribute(in_S,"REFXCDLT")) then
       xpixdelt = real64Attribute(in_S,"REFXCDLT")
    else
       if (warndie) then
          call error ("noREFXCDLT",&
               "more_coord_keys: File does not have REFXCDLT attribute")
       else
          call warning ("noREFXCDLT",&
               "more_coord_keys: File does not have REFXCDLT attribute")
       end if
    end if
    if (hasAttribute(in_A,"REFYCDLT")) then
       ypixdelt = real64Attribute(in_A,"REFYCDLT")
    elseif (hasAttribute(in_S,"REFXCDLT")) then
       ypixdelt = real64Attribute(in_S,"REFYCDLT")
    else
       if (warndie) then
          call error ("noREFYCDLT",&
               "more_coord_keys: File does not have REFYCDLT attribute")
       else
          call warning ("noREFYCDLT",&
               "more_coord_keys: File does not have REFYCDLT attribute")
       end if
    end if

    return

  end subroutine more_coord_keys
  !
  ! bhp: new sub to check and extract array data from FITS file
  !      returns 32 or 64-bin xSize by ySize array
  !      generally replaces Snowden's retfit called by:
  !      
  subroutine read_2Darray(in_S,in_A,aDR32,aDR64,aDI16,aDI32,xSize,Ysize,typeA)

    type(DataSetT), intent(inout) :: in_S
    type(ArrayT) :: in_A
    integer, dimension(:), pointer :: dimVector
    integer(kind=int16), dimension(:,:), pointer :: aDI16
    integer(kind=int32), dimension(:,:), pointer :: aDI32
    integer :: xSize, ySize, dt, typeA
    real(kind=single), dimension(:,:), pointer :: aDR32
    real(kind=double), dimension(:,:), pointer :: aDR64

    in_A = array(in_S, 0, READ) ! set already opened in check_open_set
    dimVector => dimensions(in_A)
    xSize = dimVector(0)
    ySize = dimVector(1)
    dt = dataType(in_A)
    typeA = 0
    write(errstr,*)"2D array is type: ",dt
    call message(AppMsg,Noisy)
    write(errstr,*)"2D array is dimensions (x,y): ",xSize,ySize
    call message(AppMsg,Noisy)

    SELECT CASE (dataType(in_A))
    case (INTEGER16) ! int16
       aDI16 => int16Array2Data(in_A)
       typeA = 1
    case (INTEGER32) ! int32
       aDI32 => int32Array2Data(in_A)
       typeA = 2
    case (REAL32) ! real32
       aDR32 => real32Array2Data(in_A)
       typeA = 3
    case (REAL64) ! real64
       aDR64 => real64Array2Data(in_A)
       typeA = 4
    case DEFAULT
       write(errstr,*)&
            "read_2Darray: Only I16/I32/R32/R64 out arrays supported."
       call error("badArrayType",trim(errstr))
    end SELECT
    return

  end subroutine read_2Darray
  !
  ! bhp: new sub to check and extract 3D array data from FITS file
  !      returns 32 or 64-bin xSize x ySize x zSize array
  !
  subroutine read_3Darray(in_S,aDR32,aDR64,aDI16,aDI32,xSize,Ysize,Zsize)

    type(DataSetT), intent(inout) :: in_S
    type(ArrayT) :: in_A
    integer, dimension(:), pointer :: dimVector
    integer(kind=int16), dimension(:,:,:), pointer :: aDI16
    integer(kind=int32), dimension(:,:,:), pointer :: aDI32
    integer :: xSize, ySize, zSize, dt, typeA
    real(kind=single), dimension(:,:,:), pointer :: aDR32
    real(kind=double), dimension(:,:,:), pointer :: aDR64

    in_A = array(in_S, 0, READ) ! set already opened in check_open_set
    dimVector => dimensions(in_A)
    xSize = dimVector(0)
    ySize = dimVector(1)
    zSize = dimVector(2)

    dt = dataType(in_A)
    write(errstr,*)"3D array is type: ",dt
    call message(AppMsg,Noisy)
    write(errstr,*)"3D array is dimensions (x,y,z): ",xSize,ySize,zSize
    call message(AppMsg,Noisy)

    SELECT CASE (dataType(in_A))
    case (INTEGER16) ! int16
       aDI16 => int16Array3Data(in_A)
       typeA = 1
    case (INTEGER32) ! int32
       aDI32 => int32Array3Data(in_A)
       typeA = 1
    case (REAL32) ! real32
       aDR32 => real32Array3Data(in_A)
       typeA = 3
    case (REAL64) ! real64
       aDR64 => real64Array3Data(in_A)
       typeA = 4
    case DEFAULT
       write(errstr,*)&
            "read_3Darray: Only I16/I32/R32/R64 out arrays supported."
       call error("badArrayType",trim(errstr))
    end SELECT

    return

  end subroutine read_3Darray
  !
  ! READ_HEADER_INFO: this only returns a few keywords and
  !  sanity checks/verifies file
  !                
  !
  ! bhp: partly replaces read_pha in these routines: 
  !      mosback, pnback
  !
  ! c.f. read_pha for column reading
  ! e.g. mos1U002-obj.pi
  !
  subroutine read_header_info(typeF,in_S,out_T,out_A,inst,obsid,expid,&
       smode,frmtime,filter,revl,back,area,expo,elow,ehigh,nRows,warndie)

    type(DataSetT), intent(inout) :: in_S
    type(TableT), intent(inout)   :: out_T
    type(ArrayT), intent(inout)   :: out_A
    !
    character(len=*) :: expid, obsid
    character(len=2) :: inst    
    character(len=5) :: instrume
    character(len=6) :: filter
    character(len=25) :: smode
    !
    integer :: typeF, qual, mn, readchans, inelo, inehi
    integer, intent(inout) :: nRows, elow, ehigh
    integer, intent(out) :: revl, frmtime
    integer, dimension(:), pointer :: dims
    logical, intent(inout) :: warndie
    real(kind=single), intent(out) :: back, area, expo
    back = 0.
    area = 0.
    expo = 0.
    qual = 0
    frmtime = 0
    nRows = 0
    smode = ""
    filter = ""
    inelo = elow
    inehi = ehigh
    !
    if (typeF .eq. 1) then ! event list EVLI type
       if (.not.hasBlock(in_S,'EVENTS')) then
          call error("noEVENTS","Input file does not contain an EVENTS extension.")
       else
          out_T = table(in_S,"EVENTS", READ)
          nRows = numberOfRows(out_T) ! channels or events, depending
       endif
       !
    else if (typeF .eq. 2) then
       if (.not.hasBlock(in_S,'SPECTRUM')) then
          call error("noSPECTRUM","Input file does not contain an SPECTRUM extension.")
       else
          out_T = table(in_S,'SPECTRUM', READ)
          nRows = numberOfRows(out_T) ! channels or events, depending
       endif
       !
    else if (typeF .eq. 3) then     ! no check needed for images (for now)
       out_A = array(in_S, 0, READ) ! set already opened in check_open_set
       dims => dimensions(out_A)
       write(errstr,*) "read_header_info: image file selected, dims = ",dims
       call message(AppMsg,Noisy) ! bhp: change these!!!
    else
       call error("wrongType","Input file must be event list, spectrum, or image")
    endif
    !
    ! got this far, get keywords in the primary header:
    !
    if (.not.hasAttribute(in_S,'REVOLUT')) then
       call warning("noREVOLUT",&
            "read_header_info: Input file has no REVOLUT keyword")
       revl = 0
    else
       revl = int32attribute(in_S,'REVOLUT')
    end if
    !
    if (.not.hasAttribute(in_S,'INSTRUME')) then
       call warning("noINSTRUME",&    ! only EMOS1/EMOS2/EPN are valid
            "read_header_info: Input file has no INSTRUME keyword")
       instrume = 'N/A' ! knockoff effects???!!!
    else
       instrume = stringAttribute(in_S,"INSTRUME") ! EMOS1/EMOS2/EPN
       instrume = trim(adjustl(instrume)) ! clean out whitespace
    endif
    !
    if (instrume(:3) .eq. "EMO") then  ! EPIC MOS
       inst = "M" // instrume(5:5)     ! M1 or M2
       smode = "N/A"                   ! value is for PN only
    elseif (instrume(:3) .eq. "EPN") then ! EPIC PN
       inst = "PN"
       if (.not.hasAttribute(in_S,"SUBMODE")) then
          call warning("noSUBMODE","Input file does not contain a SUBMODE keyword")
          smode = 'N/A' ! knockoff effects???!!!
       else
          smode = stringAttribute(in_S,"SUBMODE")
       endif
       if (typeF .eq. 1) then 
          if (hasAttribute(in_S,'FRMTIME')) then
             frmtime = int32Attribute(in_S,'FRMTIME')
             frmtime = 1 ! knockoff effects???!!!
          else 
             call warning("noFRMTIME",&
                  "Input EPN file does not contain a FRMTIME keyword")
          end if
       end if
    else
       print*,'instrume = ',instrume
       call error("badInfile",&
            "Input file INSTRUME is .not. EMOS1, EMOS2, or EPN")
    end if
    if (.not.hasAttribute(in_S,"ELOW")) then
       write(errstr,*) "read_header_info: No ELOW in input file"
       call message(AppMsg,Noisy) 
    else
       elow = int32Attribute(in_S,"ELOW")
    endif
    if (.not.hasAttribute(in_S,"EHIGH")) then
       write(errstr,*) "read_header_info: No EHIGH in input file"
       call message(AppMsg,Noisy)
    else
       ehigh = int32Attribute(in_S,"EHIGH")
    endif
    ! fix this!!!???
    ! user's choice still overrides what's in the file (test more for this??? SAS-20?)
!    if (inelo .ne. elow) then
!       write(errstr,*) "read_header_info: chosen elow .NE. file's elow attribute "
!       call message(AppMsg,Noisy)
!       elow = inelo ! return the user's choice
!    endif
!    if (inehi .ne. ehigh) then
!       write(errstr,*) "read_header_info: chosen ehigh .NE. file's ehigh attribute "
!       call message(AppMsg,Noisy)
!       ehigh = inehi ! return the user's choice
!    endif
    !
    !???!!! check warning vs error. will die badly later. second warndie???!!!
    if (.not.hasAttribute(in_S,"EXPIDSTR")) then ! e.g. S002 for mos or pn
       call warning("badInfile","Input file does not contain an EXPIDSTR keyword")
       expid = 'N/A' ! knockoff effects???!!!
    else
       expid = stringAttribute(in_S,"EXPIDSTR")
    endif
    if (.not.hasAttribute(in_S,"FILTER")) then ! e.g. S002 for mos or pn
       call warning("badInfile","Input file does not contain an FILTER keyword")
       filter = 'N/A' ! knockoff effects???!!!
    else
       filter = stringAttribute(in_S,"FILTER")
    endif
    if (.not.hasAttribute(in_S,'OBS_ID')) then  
       call warning("noOBS_ID",&
            "read_header_info: Input file has no OBS_ID keyword")
       obsid = 'N/A' ! knockoff effects???!!!
    else
       obsid = stringAttribute(in_S,'OBS_ID') 
    end if
    !
    back = 1.
    area = 1.
    expo = 1.
    !
    if (typeF .eq. 2) then
       if (.not. hasAttribute(in_S,'BACKSCAL')) then
          if (hasAttribute(out_T,'BACKSCAL')) then
             back = real32attribute(out_T,'BACKSCAL')
          else
             call warning ("noBACKSCAL",&
                  "No BACKSCAL attribute found in file, setting to 1.")
          end if
       else
          back = real32attribute(in_S,'BACKSCAL')
       endif
       if (back .eq. 0.) then
          call warning ("zeroBACKSCAL",&
               "BACKSCAL attribute found in file, but it is 0.!")
       end if
       !
       if (.not. hasAttribute(in_S,'AREASCAL')) then
          if (hasAttribute(out_T,'AREASCAL')) then
             area = real32attribute(out_T,'AREASCAL')
          else
             call warning ("noAREASCAL",&
                  "No AREASCAL attribute found in file, setting to 1.")
          end if
       else
          area = real32attribute(in_S,'AREASCAL')
       endif
       if (area .eq. 0.) then
          call warning ("zeroAREASCAL",&
               "AREASCAL attribute found in file, but it is 0.!")
       end if
       !
       if (.not. hasAttribute(in_S,'EXPOSURE')) then
          if (hasAttribute(out_T,'EXPOSURE')) then
             expo = real32attribute(out_T,'EXPOSURE')
          else
             call warning ("noEXPOSURE",&
                  "No EXPOSURE attribute found in file, setting to 1.")
          end if
       else
          expo = real32attribute(in_S,'EXPOSURE')
       endif
       if ( expo .eq. 0.) then
          call error ("zeroEXPOSURE",&
               "EXPOSURE attribute found in file, but it is 0.!")
       end if
    end if
    ! why doesn't Noisy seem to work???!!!
!    write(errstr,*) &
!         "read_header_info: BACKSCAL/AREASCAL/EXPOSURE:"
!    call message(AppMsg,Noisy)
!    write(errstr,*) &
!         "backscal/areascal/exposure/quality: ",back,"/",area,"/",expo,"/",qual
!    call message(AppMsg,Noisy)
!    write(errstr,*) "read_header_info: OBSID=",trim(obsid)," EXPID=",trim(expid),&
!         " INSTRUME=",inst," SUBMODE(PN)=",trim(adjustl(smode))," FILTER=",trim(adjustl(filter))
!    call message(AppMsg,Noisy)

    return

  end subroutine read_header_info
  !
  ! READ_PHA: read in XMM-SAS Spectral (PHA) File,
  ! e.g. mos1U002-obj.pi
  !
  subroutine read_pha (in_T,inchans,type,chan_ptr,j_ptr,r_ptr,re_ptr)
    !
    ! SAS OAL datatypes:
    !
    type(TableT), intent(inout)  :: in_T
    type(ColumnT) :: in_C    
    !
    character(len=8), parameter :: columns(4) = &
         (/'CHANNEL ','COUNTS  ','GROUPING','QUALITY '/)
    character(len=8), parameter :: extname = "SPECTRUM"

    integer(kind=int16), dimension(:), pointer :: chan_ptr
    integer(kind=int32), dimension(:), pointer :: j_ptr
    integer(kind=int32) :: i, type, inchans
    integer(kind=int16), allocatable, dimension(:) :: channels
    integer(kind=int32), allocatable, dimension(:) :: &
         grouping, quality ! bhp: not needed, but keep for now
    real(kind=single), dimension(:), pointer :: r_ptr, re_ptr
    !
    ! bhp: set already checked and opened in check_open_set subroutine,
    !      extract attributes (keywords) from SPECTRUM extension
    !
    ! bhp: read in all the PHA FITS columns
    !      pha MUST have CHANNEL and COUNTS columns
    !
    ! inchan is already numberOfrows from check_open_set
    ! channels is the array containing all the channel numbers
    write(errstr,*) "GOFIO start."
    call message(AppMsg,Noisy)
    !
    in_C = column(in_T,"CHANNEL", READ)
    chan_ptr => int16Data(in_C)
    if ( .not. hasColumn(in_T,'CHANNEL')) then 
       write(errstr,*)"Missing CHANNEL column in PHA file"
       call error("noCHANNELcol",trim(errstr))
    end if
    !
    if ( .not. hasColumn(in_T,'COUNTS')) then 
       write(errstr,*)"Missing COUNTS column in PHA file"
       call error("noCOUNTScol",trim(errstr))
    else
       if (type .eq. 1) then ! counts are i32
          in_C = column(in_T,"COUNTS", READ) 
          j_ptr => int32Data(in_C)
          call release(in_C)
          if (hasColumn(in_T,'STAT_ERR')) then 
             in_C = column(in_T,"STAT_ERR", READ)
             re_ptr => real32Data(in_C) ! check this ???!!!
             call release(in_C)
          else
             allocate(re_ptr(0:SIZE(j_ptr)-1))
             do i=0,inchans-1
                re_ptr(i) = sqrt(abs(real(j_ptr(i))))
             end do
          end if
       elseif(type .eq. 2) then
          in_C = column(in_T,"COUNTS", READ) 
          r_ptr => real32Data(in_C)
          call release(in_C)
          if (hasColumn(in_T,'STAT_ERR')) then 
             in_C = column(in_T,"STAT_ERR", Read)
             re_ptr = real32Data(in_C)
             call release(in_C)
          else
             allocate(re_ptr(0:SIZE(j_ptr)-1))
             do i=0,inchans-1
                re_ptr(i) = sqrt(abs(real(j_ptr(i))))
             end do
          end if
       elseif (type .eq. 0) then

       else
          call error("wrongTYPE",&
               "spec file has neither J or E type COUNTS column")
       end if
    end if
    return

  end subroutine read_pha
  !
  ! READ_RMF: read in XMM-SAS Redistribution Matrix File (RMF)
  !
  ! old: subroutine read_rmf (obj_rmf,mchan,detm,chan,eminin,emaxin,nchanrin,&
  !       eminout,emaxout,nchanrout)

  !
  ! READ_RMF: read in XMM-SAS OGIP-compliant Response MATRIX File (RMF)
  ! used in: mosback, pnback, swcx
  ! 
  subroutine read_rmf (in_S,inChans,rmfChans,emin,emax,matrix)
    !
    ! SAS OAL datatypes:
    !
    type(DataSetT), intent(inout) :: in_S
    type(TableT)   :: in_T
    type(ColumnT)  :: in_C    
    !
    character(len=8), parameter :: columns(3) = &
         (/'CHANNEL ', 'E_MIN   ','E_MAX   '/)
    character(len=8), parameter :: extname = "EBOUNDS"

    integer :: inChans, rmfChans, i, j, matrixSize, nElems
    integer :: chanArr(inChans)
    integer(int16), dimension(:), pointer :: i_ptr
    integer(int32), dimension(:), pointer :: j_ptr
    integer,dimension(:),pointer :: dimCol

    real(kind=single), dimension(:), pointer :: r_ptr
    real(kind=single), dimension(:), pointer :: matrix_ptr
    real(kind=single), intent(inout) :: emin(inChans), emax(inChans)
    real(kind=single), dimension(:,:), allocatable :: matrix
    real(kind=single), dimension(:), allocatable :: detmt
    !
    ! bhp: set already checked and opened in check_open_set subroutine,
    !      extract attributes (keywords) from EBOUNDS extension
    ! sanity check first
    if (.not.hasBlock(in_S,'MATRIX')) then  
       call error("noMATRIX",&
            "read_rmf: Input file has no MATRIX extension")
    end if
    if (.not.hasBlock(in_S,'EBOUNDS')) then  
       call error("noEBOUNDS",&
            "read_rmf: Input file has no EBOUNDS extension")
    else
       in_T = table(in_S,'MATRIX')
       matrixSize = numberOfRows(in_T)
       allocate(matrix(matrixSize,matrixSize))
       if (.not. hasColumn(in_T,'MATRIX')) then
          write(errstr,*)"Missing MATRIX column in RMF file"
          call error("noMATRIX",trim(errstr))
       else
          in_C = column(in_T,'MATRIX',Read)
          nElems = numberOfElements(in_C)
          dimCol => dimensions(in_C)
          do i=0,matrixSize-1
             allocate (detmt(numberOfElements(in_C, i)))
             matrix_ptr => real32Data(in_C,i)
             do j=0,numberOfElements(in_C, i)-1
                matrix(i,j) = matrix_ptr(j)
             end do
             deallocate(detmt)
          end do
       end if
    end if
    call release(in_T)
    call release(in_C)
    in_T = table(in_S,'EBOUNDS', READ)
    rmfChans = numberOfRows(in_T)
    ! bhp: should this be fatal???? I say yes, because it has to
    ! be the same size as the obj spec allocated space
    if (rmfChans .ne. inChans) then  
       call warning("inChansNErmfChans",&
            "Spec file nChans is not equal to chans in RMF")
    end if
    ! bhp: read in all the PHA FITS columns
    !      pha MUST have CHANNEL and COUNTS columns
    if ( .not. hasColumn(in_T,'CHANNEL')) then 
       write(errstr,*)"Missing CHANNEL column in RMF file"
       call error("noCHANNEL",trim(errstr))
    else
       in_C = column(in_T,"CHANNEL", Read) 
       i_ptr => int16Data(in_C)
       chanArr(1:rmfChans) = int(i_ptr(0:rmfChans-1))
       call release(in_C)
    end if
    if ( .not. hasColumn(in_T,'E_MIN')) then 
       write(errstr,*)"Missing E_MIN column in RMF file"
       call error("noEMIN",trim(errstr))
    else
       in_C = column(in_T,"E_MIN", Read) 
       r_ptr => real32Data(in_C)
       emin(1:rmfChans) = real(r_ptr(0:rmfChans-1))
       call release(in_C)
    end if
    if ( .not. hasColumn(in_T,'E_MAX')) then 
       write(errstr,*)"Missing E_MAX column in RMF file"
       call error("noEMAX",trim(errstr))
    else
       in_C = column(in_T,"E_MAX", Read) 
       r_ptr => real32Data(in_C)
       emax(1:rmfChans) = real(r_ptr(0:rmfChans-1))
       call release(in_C)
    end if
    !
    call release(in_T)

    return

  end subroutine read_rmf
  !
  ! READ_ARF: read in XMM-SAS OGIP-compliant Ancillary Response File (ARF)
  ! used ONLY in: swcx_mod.f90
  ! 
  subroutine read_arf (in_S,inChans,outChans,elow,ehigh,arf)

    type(DataSetT) :: in_S
    type(TableT)   :: in_T
    type(ColumnT)  :: in_C    

    character (len=8),parameter :: extname = 'SPECRESP'
    character(len=8), parameter :: columns(3) = &
         (/'ENERG_LO', 'ENERG_HI','SPECRESP'/)
    integer, intent(in) :: inChans
    integer :: outChans, i 
    integer(int16), dimension(:), pointer :: i_ptr
    integer(int32), dimension(:), pointer :: j_ptr
    real, intent(inout) :: elow(inChans), ehigh(inChans), arf(inChans)
    real(kind=single), dimension(:), pointer :: r_ptr

    !  Read the FITS file header information


    if ( hasBlock(in_S,trim(extname)) ) then
       in_T = table(in_S,extname)
    else
       write(errstr,*)"ARF does NOT contain SPECRESP block"
       call error("noSPECRESP",trim(errstr))
    end if
    outChans = numberOfRows(in_T)
    if (outChans .le. 0) then
       call error("NoRows",&
            "ARF SPECRESP is empty (0 chans)")
    end if

    !  ss: Read in the column names
    ! bhp: but in the 'new' way (and if any are missing, die gracefully)

    do i=1,size(columns) ! bhp: 3 in this case
       if ( .not. hasColumn(in_T,trim(columns(i))) ) then 
          write(errstr,*)"Input file column: ",&
               trim(columns(i))," missing from ARF file"
          call error("missingColumn",trim(errstr))
       end if
    end do
    !  ss:  Extract the channel numbers
    ! bhp: read arf FITS table

    in_C = column(in_T,"ENERG_LO", Read) 
    r_ptr => real32Data(in_C)
    elow(1:outChans)  = r_ptr(0:outChans-1)
    call release(in_C)
    in_C = column(in_T,"ENERG_HI", Read) 
    r_ptr => real32Data(in_C)
    ehigh(1:outChans)  = r_ptr(0:outChans-1)
    call release(in_C)
    in_C = column(in_T,"SPECRESP", Read) 
    r_ptr => real32Data(in_C)
    arf(1:outChans)  = r_ptr(0:outChans-1)
    call release(in_C)    

    return

  end subroutine read_arf
  !
  ! generally replaces subroutine putfit
  ! writes created 2D arrays to new file
  ! called by: adapt_merge, adapt, bin_image_merge, bin_image,
  !            merge_comp_xmm, merge_off_xmm
  !
  subroutine write_2Darray(out_S,type,name,i16array,i32array,r32array,&
       r64array,xSize,ySize,crpix,crval,cdelt,cunit,comment,nc)

    type(DataSetT) :: out_S ! bhp: set already opened with check_open_set
    type(ArrayT) :: out_A
    integer, intent(in) :: xSize, ySize, nc

    character(len=8), intent(in) :: cunit(2), name
    character(len=80), intent(in) :: comment(nc)

    integer :: i, j, type ! 1=i16, 2=i32, 3=r32, 4=r64
    integer(kind=int16), intent(in) :: i16array(xSize,ySize)
    integer(kind=int32), intent(in) :: i32array(xSize,ySize)
    integer, dimension(:,:), pointer :: i32_P
    integer(kind=int16), dimension(:,:), pointer :: i16_P

    real(kind=single), intent(in) ::  r32array(xSize,ySize)
    real(kind=single), dimension(:,:), pointer :: r32_P
    real(kind=double), dimension(:,:), pointer :: r64_P
    real(kind=double), intent(in) :: r64array(xSize,ySize),&
         crpix(2), crval(2), cdelt(2)

    SELECT CASE (type)
    case (1) ! int16
       out_A = addArray(out_S, trim(name), INT16, (/xSize,ySize/))
       i16_P => int16Array2Data(out_A)
       i16_P(0:xSize-1,0:ySize-1) = i16array(1:xSize,1:ySize)
    case (2) ! int32
       out_A = addArray(out_S, trim(name), INT32, (/xSize,ySize/))
       i32_P => int32Array2Data(out_A)
       i32_P(0:xSize-1,0:ySize-1) = i32array(1:xSize,1:ySize)
    case (3) ! real32
       out_A = addArray(out_S, trim(name), REAL32, (/xSize,ySize/))
       r32_P => real32Array2Data(out_A)
       r32_P(0:xSize-1,0:ySize-1) = r32array(1:xSize,1:ySize)
    case (4) ! real64
       out_A = addArray(out_S, trim(name), REAL64, (/xSize,ySize/))
       r64_P => real64Array2Data(out_A)
       r64_P(0:xSize-1,0:ySize-1) = r64array(1:xSize,1:ySize)
    case DEFAULT
       write(errstr,*)&
            "write_2Darray: Only I16/I32/R32/R64 out arrays supported."
       call error("badArrayType",trim(errstr))
    end SELECT

    ! bhp: add some keywords and whatnot

    call setAttribute(out_S,"TELESCOP","XMM","Mission") 
    call setAttribute(out_S,"INSTRUME","EPIC","Instrument name") 
    call setAttribute(out_S,"OBS_MODE","POINTING","Observation mode (pointing/slew/scan)") 
    call setAttribute(out_S,"RADECSYS","FK5","Equatorial system reference") 
    call setAttribute(out_S,"EQUINOX",2000.,"Equinox") 
    call setAttribute(out_S,"BZERO",0.,"Zero offset")   ! bhp: always???
    call setAttribute(out_S,"BSCALE",1.,"Scale factor") ! bhp: always???
    call setAttribute(out_S,"BUNIT",'c/s/deg2',"Pixel units")

    call setAttribute(out_S,"CRPIX1",crpix(1),"Reference X Pixel")
    call setAttribute(out_S,"CRPIX2",crpix(2),"Reference Y Pixel")
    call setAttribute(out_S,"CTYPE1","RA---TAN","Projection")
    call setAttribute(out_S,"CTYPE2","DEC--TAN","Projection")
    call setAttribute(out_S,"CRVAL1",crval(1),"Reference coordinate for first axis (RA)")
    call setAttribute(out_S,"CRVAL2",crval(2),"Reference coordinate for second axis (DEC)")
    call setAttribute(out_S,"CDELT1",cdelt(1),"Pixel X size")
    call setAttribute(out_S,"CDELT2",cdelt(2),"Pixel Y size")
    call setAttribute(out_S,"CUNIT1",cunit(1),"Units of coordinate")
    call setAttribute(out_S,"CUNIT2",cunit(2),"Units of coordinate")
    do i = 1, nc
       call addComment(out_S,trim(comment(i)))
    end do

    call release(out_A)

    return

  end subroutine write_2Darray
  !
  ! OPEN_ASCII_FILE: Check/open, e.g. qdp file for write
  !
  subroutine open_ascii_file(filename,lunit)

    character(*), intent(in) :: filename
    integer, intent(out) :: lunit
    integer :: status
    logical :: exists

    lunit = GetLun()

    inquire(FILE=trim(filename),EXIST=exists)
    
    if( clobber() ) then ! use SAS_CLOBBER
       if (exists) then 
          open(unit=lunit,status='replace',form='formatted',&
               file=trim(filename),iostat=status)
       else
          open(unit=lunit,status='new',form='formatted',&
               file=trim(filename),iostat=status)
       end if
    else
       if (exists) then 
          call error ("noClobber","File exists and SAS_CLOBBER is NOT set")
       else
          open(unit=lunit,status='new',form='formatted',&
               file=trim(filename),iostat=status)
       endif

    endif

    return

  end subroutine open_ascii_file
  !
  !
  !
  subroutine parse_prefix(string,inst,soru,expid,prefix)

    ! used by: mos_back, pn_back
    character(len=*), intent(in) :: string ! filename OR prefix
    character(len=1), intent(out) :: soru
    character(len=3), intent(out) :: expid
    character(len=4), intent(out) :: inst
    character(len=5), intent(out) :: prefix
    character(len=10) :: obsid !bhp: should have intent later!!!

    ! if old style, string = 1S002/1U005/S005

    if (len(string) .eq. 4) then ! PN only
       inst = 'pn'
       soru = string(:1)
       expid = string (2:)
       prefix = trim(string)
    elseif (len(string) .eq. 5) then ! MOS only
       inst = 'mos' 
       soru = string(2:2)
       expid = string(3:5)
       prefix = trim(string)
    else 
       ! if new style, string = P0097820101M1S005-objimg.FIT
       ! if new style, string = P0097820101PNS005-objimg.FIT
       if (string(12:12) .eq. "M") then 
          inst = 'mos' // string(13:13)
          prefix = string(13:17)
       else 
          inst = 'pn'
          prefix = string(14:17)
       end if
       soru  = string(14:14)
       expid = string(15:17)
       obsid = string(2:11)
    end if
    return

  end subroutine parse_prefix

  subroutine make_filename (inst,expid,type,elow,ehigh,suffix,filename,withlong)

    character(len=*) :: inst, expid, type, suffix
    character(len=6) :: prefix="",clow="",chigh=""
    character(len=15) :: e_string=""
    character(len=*), intent(out) :: filename
    integer, intent(in) :: elow, ehigh
    logical :: withlong

    prefix = adjustl(trim(inst)) // adjustl(trim(expid)) ! e.g. M1S001
    ! only used if needed
    ! e_string max = -10000-10000-
    if (withlong) then 
       if ((elow .ne. 0) .and. (ehigh .ne. 0)) then
          if ((elow .gt. 32767) .or. (ehigh .gt. 32767)) then
             call error("elowORehighTooBig","Max value for elow or ehigh is 32767")
          end if
          write(clow,"(I5)") elow
          write(chigh,"(I5)") ehigh
          e_string = "-" // trim(adjustl(clow)) // "-" // trim(adjustl(chigh))
       endif
    endif
    !
    if ((inst .eq. 'M1') .or. (inst .eq. 'M2')) then
       filename = "mos" // trim(adjustl(prefix(2:))) // "-" // trim(adjustl(type)) // &
            trim(adjustl(e_string)) // "." // trim(adjustl(suffix))
    else
       filename = "pn" // trim(adjustl(prefix(3:))) // "-" // trim(adjustl(type)) // &
            trim(adjustl(e_string)) // "." // trim(adjustl(suffix))
    endif

    return

  end subroutine make_filename
  !
  !
  !
  subroutine addtask(in_F,task)
    !
    ! add taskname to a file's primary header 
    ! 4 = 'open MODIFY'
    ! warndie = fail if can't add attribute
    !
    type(DataSetT) :: in_S
    character (len=*) :: in_F, task
    character (len=80) :: comment
    integer :: status=0
    logical :: warndie = .TRUE.
    call check_open_set(trim(in_F),in_S,4,status,warndie)
    comment = "This file was created within the task: " // trim(adjustl(task))
    call addComment(in_S,trim(adjustl(task)))
    call release(in_S)
    return
  end subroutine addtask

  !
  !
  !
  subroutine addlohi(in_F,elow,ehigh)
    !
    ! add elow/ehigh to a file's primary header 
    ! 4 = 'open MODIFY'
    ! warndie = fail if can't add attribute
    !
    type(DataSetT) :: in_S
    character (len=*) :: in_F
    integer :: status=0, elow, ehigh
    logical :: warndie = .TRUE.    
    if ((elow .gt. 0) .and. (ehigh .gt. 0)) then
       call check_open_set(trim(in_F),in_S,4,status,warndie)
       call setAttribute(in_S,"ELOW" ,elow,"eV","Min energy channel")
       call setAttribute(in_S,"EHIGH",ehigh,"eV","Max energy channel") 
       call release(in_S)
    end if
    return
  end subroutine addlohi
  !
  !
  !
  subroutine addpattern(in_F,pattern)
    !
    ! add elow/ehigh to a file's primary header 
    ! 4 = 'open MODIFY'
    ! warndie = fail if can't add attribute
    !
    type(DataSetT) :: in_S
    character (len=*) :: in_F
    integer :: status=0, pattern
    logical :: warndie = .TRUE.    
    call check_open_set(trim(in_F),in_S,4,status,warndie) ! 4=MODIFY
    call setAttribute(in_S,"PATTERN" ,pattern,"","evselect PATTERN")
    call release(in_S)
    return
  end subroutine addpattern
  !
  !
  !
  subroutine check_clobber(outfile)

    character(len=*) :: outfile
    logical :: exists
    !
    ! In any ESAS task, if clobber is not set, die (early!)
    !
    inquire(FILE=trim(outfile),EXIST=exists)
    if (exists) then
       if (clobber()) then
          write(errstr,*)"Output file, " // trim(adjustl(outfile)) // &
               " exists and clobber IS set. Deleting file, creating new version."
          call message(AppMsg,Always) ! change to Noisy???!!!
          call unlink (trim(outfile))
       else
          write(errstr,*)"Output file, " // trim(adjustl(outfile)) // &
               " exists and clobber is NOT set, exiting."
          call error("noClobber",trim(adjustl(errstr)))
       endif
    end if

    return
    
  end subroutine check_clobber
  
  !
  ! fill other arrays
  ! These have 900x900 Real32 arrays
  !
  subroutine getotherim(type,nprefixes,prefixes,elow,ehigh,mask_RP)

    type(DataSetT) :: mask_S
    type(ArrayT) :: mask_A
    !
    character(len=*) :: prefixes(nprefixes)
    character(len=256) :: mask_F
    character(len=12) :: addon
    character(len=5) :: celow, cehigh, currprefix
    character(len=3) :: inst, type
    integer :: typeA, xSize=900, ySize=900
    integer, dimension(:), pointer :: dimVector
    integer(kind=int16), dimension(:,:), pointer :: aDI16
    integer(kind=int32), dimension(:,:), pointer :: aDI32
    integer :: i, j, k, nprefixes, status, elow(nprefixes), ehigh(nprefixes)
    integer, dimension(:,:), pointer :: mask_IP
    logical :: warndie
    real(kind=single), intent(inout), dimension(:,:,:), pointer :: mask_RP
    real(kind=single), dimension(:,:), pointer :: inMask_RP
    real(kind=double), dimension(:,:), pointer :: aDR64
    !
    mask_RP = 1.
    do i=1,nprefixes 
       inMask_RP = 0. ! or 1. ???!!!
       currprefix = prefixes(i)
       if (currprefix(1:1) .eq. '1') then
          inst = 'mos'
       elseif (currprefix(1:1) .eq. '2') then
          inst = 'mos'
       else
          inst = 'pn'
       endif
       !
       write(celow,'(I5)')elow(i)
       write(cehigh,'(I5)')ehigh(i)
       select case (type)
       case ('swc')
          addon = "-swcximsky-"
       case ('qpb')
          addon = "-backimsky-"
       case('pro')
          addon = "-protimsky-"
       case default
          call error("noType","Type is not back, prot, or swcx")
       end select
       !
       mask_F = trim(adjustl(inst)) // trim(adjustl(prefixes(i))) // &
            trim(adjustl(addon)) // trim(adjustl(celow)) // "-" // &
            trim(adjustl(cehigh)) // ".fits"
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_open_set(mask_F,mask_S,1,status,warndie) ! goflib/gofio_mod
       call read_2Darray(mask_S,mask_A,inMask_RP,aDR64,aDI16,aDI32,xSize,Ysize,typeA)
       do j = 0,xSize-1
          do k = 0,ySize-1
             mask_RP(j,k,i) = inMask_RP(j,k)
          enddo
       enddo
       call release(mask_S)
    enddo

    return

  end subroutine getotherim
  !
  ! eliminate ccd selected if occurring after CCD3/6 loss
  !
  subroutine check_orbit(ccds,rev)
    character(len=1) :: cccd
    integer :: rev, i
    logical, intent(inout) :: ccds(7)
    !
    ! BHP: the below is ordinarily handled by the DAL/CAL but with the FWC
    !      data it still needs to be done or there's a seg fault if a user
    !      selects these conditions.
    !
    write(errstr,*)"Checking if observation took place after MOS1 CCD3 or CCD6 was lost"
    call message(AppMsg,Always) ! change to Noisy!!!??? 
    if (ccds(6)) then
       if (rev .gt. 961) then
          write(errstr,*)"Observation took place after MOS1 CCD6 was lost"
          call message(AppMsg,Always) ! change to Noisy!!!??? 
          write(errstr,*)"Removing CCD6 from user-selected ccds."
          call message(AppMsg,Always) ! change to Noisy!!!???
          ccds(6) = .FALSE.
       elseif(rev .eq. 961) then
          write(errstr,*)'Observation took place during the exact orbit where ',&
               'MOS1 CCD6 was lost. Exposure U002 has CCD6 data, U003 does not'
          call warning("Rev961",trim(adjustl(errstr)))
       end if
    end if
    if (ccds(3)) then
       if (rev .gt. 2382) then
          write(errstr,*)"Observation took place after MOS1 CCD3 was lost"
          call message(AppMsg,Always) ! change to Noisy!!!??? 
          write(errstr,*)"Removing CCD3 from user-selected ccds."
          call message(AppMsg,Always) ! change to Noisy!!!???
          ccds(3) = .FALSE.
       end if
    end if
    write(errstr,*)"Updated CCD selection:"
    call message(AppMsg,Always) ! change to Noisy!!!???
    do i=1,7
       write(cccd,'(I1)')i
       write(errstr,*) "CCD(",cccd,") selected? (y/n)",ccds(i)
       call message(AppMsg,Always)
    end do
    return
    
  end subroutine check_orbit
  !
  !
  !
  subroutine retfit(infile,rarray,i4array, &
       iflag,isize1,isize2,ctype,crval,cdelt,crpix,status)
    !
    !  INFILE    file name of input file
    !  RARRAY    REAL*4 array for input
    !  I4ARRAY   INTEGER*4 array for input
    !  IFLAG     flag for input file type selection
    !            1 => R*4 array
    !            2 => I*4 array
    !  ISIZE1    array dimension 1
    !  ISIZE2    array dimension 2
    !  CRVAL1    reference coordinate for first axis (RA)
    !  CDELT1    pixel size for first axis
    !  CRVAL2    reference coordinate for second axis (Dec)
    !  CDELT2    pixel size for second axis
    !  STATUS    FITSIO status flag

    implicit none

    integer :: isize1, isize2
    integer :: bitpix, blockn, hdutype, &
         i4array(isize1,isize2), iflag, &
         gcount, naxes(10), naxis, lun, pcount, status
    real :: rarray(isize1,isize2)
    character(len=8) :: ctype(2)
    character(len=80) :: com
    character(len=132) :: infile
    double precision :: cdelt(2), crpix(2), crval(2)
    logical anyf, extend, simple
    integer :: err = 0

    status = 0

    !  Get an unused lun

    call ftgiou(lun,status)
    if((status .ne. 0) .and. (err .eq. 0)) then
       err = 1
       print *, 'RETFIT POSITION 1'
       print *, 'STATUS FTGIOU =', status
    endif

    !  Read the array

    write(errstr,*) "RETFIT file name: ", trim(infile)
    call message(AppMsg,Noisy)

    !  Open the FITS file for the first observation exposure

    call ftopen(lun,infile,0,blockn,status)
    if((status .ne. 0) .and. (err .eq. 0)) then
       print *, 'RETFIT POSITION 2'
       print *, 'STATUS FTOPEN =', status
       print *, infile
       call ftclos (lun,status)
       call ftfiou (lun,status)
       return
    endif

    !  Read the FITS file header information

    call ftghpr (lun,10,simple,bitpix,naxis,naxes, &
         pcount,gcount,extend,status)

    !  Check to see if the primary has the image

    if(naxis .eq. 0) then
       call ftmrhd (lun,1,hdutype,status)
       call ftghpr (lun,10,simple,bitpix,naxis,naxes, &
            pcount,gcount,extend,status)
    endif

    if((isize1 .ne. naxes(1)) .or. (isize2 .ne. naxes(2))) then
       print *, 'RETFIT POSITION 3a, STATUS = ', STATUS
       print *, infile
       print *, 'Input array size not equal to image'
       print *, isize1, isize2, ' --  ', naxes(1), naxes(2)
       stop
    endif
    if((status .ne. 0) .and. (err .eq. 0)) then
       err = 1
       print *, 'RETFIT POSITION 3b'
       print *, 'STATUS FTGHPR =', status
       print *, infile
       stop
    endif

    !  Get the useful keywords

    call ftgkys(lun,'CTYPE1',ctype(1),com,status)
    call ftgkyd(lun,'CRVAL1',crval(1),com,status)
    if(status .ne. 0) then
       print *, 'RETFIT failure on CTYPE1'
       stop
    endif
    call ftgkyd(lun,'cdelt1',cdelt(1),com,status)
    if(status .ne. 0) then
       print *, 'RETFIT failure on CDELT1'
       stop
    endif
    call ftgkyd(lun,'crpix1',crpix(1),com,status)
    if(status .ne. 0) then
       print *, 'RETFIT failure on CRPIX1'
       stop
    endif
    call ftgkys(lun,'CTYPE2',ctype(2),com,status)
    call ftgkyd(lun,'CRVAL2',crval(2),com,status)
    if(status .ne. 0) then
       print *, 'RETFIT failure on CRVAL2'
       stop
    endif
    call ftgkyd(lun,'CDELT2',cdelt(2),com,status)
    if(status .ne. 0) then
       print *, 'RETFIT failure on CDELT2'
       stop
    endif
    call ftgkyd(lun,'CRPIX2',crpix(2),com,status)
    if(status .ne. 0) then
       print *, 'RETFIT failure on CRPIX2'
       stop
    endif

    !  Read in an R*4 file

    if(iflag .eq. 1) then
       call ftg2de (lun,0,0,naxes(1),naxes(1),naxes(2),rarray, &
            anyf,status)
       if((status .ne. 0) .and. (err .eq. 0)) then
          err = 1
          print *, 'RETFIT POSITION 4'
          print *, 'STATUS FTG2DE =', status
          print *, infile
       endif

       !  Read in an I*4 file

    elseif(iflag .eq. 2) then
       call ftg2dj (lun,0,0,naxes(1),naxes(1),naxes(2),i4array, &
            anyf,status)
       if((status .ne. 0) .and. (err .eq. 0)) then
          err = 1
          print *, 'RETFIT POSITION 6'
          print *, 'STATUS FTG2DJ =', status
          print *, infile
       endif
    endif

    !  Close the FITS file

    call ftclos(lun,status)
    if((status .ne. 0) .and. (err .eq. 0)) then
       err = 1
       print *, 'RETFIT POSITION 7'
       print *, 'STATUS FTCLOS =', status
       print *, infile
    endif

    !  Clean up

    status = 0
    call ftfiou (lun,status)

    return
  end subroutine retfit
  !
  !
  !
  subroutine putfit(infile,rarray,i4array,iflag, &
       units,comm,isize1,isize2,bzero,bscale,ctype,crval,cdelt, &
       crpix,clobber,status)

    !  INFILE    file name of output file
    !  RARRAY    REAL*4 array for output
    !  I4ARRAY   INTEGER*4 array for output
    !  IFLAG     flag for input file type selection
    !            1 => R*4 array
    !            2 => I*4 array
    !  UNITS     character string showing the units of the array
    !  COMM      comment character string
    !  ISIZE     array (square) dimension
    !  BZERO     offset value for array
    !  BSCALE    slope value for array
    !  CRVAL1    reference coordinate for first axis (RA)
    !  CDELT1    pixel size for first axis
    !  CRVAL2    reference coordinate for second axis (Dec)
    !  CDELT2    pixel size for second axis
    !  CLOBBER   1 = overwrite an existing file
    !            0 = don't
    !  STATUS    FITSIO status flag

    implicit none

    integer :: isize1, isize2
    real :: bzero, bscale, cdelt1, cdelt2, &
         crpix1, crpix2, crval1, crval2, rarray(isize1,isize2)
    double precision :: cdelt(2), crpix(2), crval(2)
    integer :: i4array(isize1,isize2)
    character(len=5) :: com
    character(len=8) :: ctype(2), units
    character(len=60) :: comm, comment
    character(len=132) :: infile
    integer :: bitpix, blockn, fpixel, iflag, &
         gcount, group, naxes(10), naxis, lun, pcount, status
    logical :: clobber, extend, simple

    equivalence (comment,com)

    integer :: err = 0

    status = 0

    !  Get an unused lun

    call ftgiou(lun,status)
    if((status .ne. 0) .and. (err .eq. 0)) then
       err = 1
       print *, 'STATUS FTGIOU =', status
       print *, 'POSITION 31'
    endif
    comment = comm

    !  Write out the array

    write(errstr,*) "PUTFIT file name: ", trim(infile)
    call message(AppMsg,Noisy)

    simple = .TRUE.
    naxis = 2
    naxes(1) = isize1
    naxes(2) = isize2
    pcount = 0
    gcount = 1
    extend = .FALSE.
    if(iflag .eq. 1) then
       bitpix = -32
    elseif(iflag .eq. 2) then
       bitpix = 16
    elseif(iflag .eq. 3) then
       bitpix = 32
    endif
    group = 1
    fpixel = 1

    !  Check to see if there is an existing file

    call ftopen (lun,infile,0,blockn,status)
    if(status .eq.  0) then
       print *, 'File exists with output file name'
       if(clobber) then
          print *, 'Clobber the file and continue'
          call ftdelt(lun,status)
       else
          print *, 'Will not clobber file'
          print *, 'XMM-BACK-IMAGE POSITION 0'
          stop
       endif
    endif
    status = 0

    !  Open the output file

    call ftinit(lun,infile,2880,status)
    if((status .ne. 0) .and. (err .eq. 0)) then
       err = 1
       print *, 'STATUS FTINIT =', status
       print *, 'POSITION 1'
       print *, infile
    endif

    !  Set the output file header

    call ftphpr(lun,simple,bitpix,naxis,naxes,pcount,gcount, &
         extend,status)
    if((status .ne. 0) .and. (err .eq. 0)) then
       err = 1
       print *, 'STATUS FTPHPR =', status
       print *, 'POSITION 2'
       print *, infile
    endif

    !  Write some useful header keywords

    call ftpcom(lun,comment,status)
    call ftpkys(lun,'CONTENT','IMAGE','  ',status)
    call ftpkys(lun,'TELESCOP','XMM','mission name',status)
    call ftpkys(lun,'INSTRUME','EPIC','instrument name',status)
    call ftpkys(lun,'OBS_MODE','POINTING', &
         'obs mode: POINTING,SLEW, OR SCAN',status)
    call ftpkys(lun,'RADECSYS','FK5', &
         'Equatorial system reference',status)
    call ftpkyj(lun,'EQUINOX',2000,'Equinox',status)
    call ftpkyf(lun,'BZERO',bzero,4,' ',status)
    call ftpkyf(lun,'BSCALE',bscale,8,' ',status)
    call ftpkys(lun,'BUNIT',units,'Units of data',status)
    crpix1 = crpix(1)
    call ftpkyf(lun,'CRPIX1',crpix1,4,'Reference pixel',status)
    call ftpkys(lun,'CTYPE1',ctype(1),'Projection',status)
    crval1 = crval(1)
    call ftpkyf(lun,'CRVAL1',crval1,4,'Central longitude',status)
    cdelt1 = cdelt(1)
    call ftpkyf(lun,'CDELT1',cdelt1,8,'Pixel size',status)
    call ftpkys(lun,'CUNIT1','deg','Units of coordinate',status)

    crpix2 = crpix(2)
    call ftpkyf(lun,'CRPIX2',crpix2,4,'Reference pixel',status)
    call ftpkys(lun,'CTYPE2',ctype(2),'Projection',status)
    crval2 = crval(2)
    call ftpkyf(lun,'CRVAL2',crval2,4,'Central latitude',status)
    cdelt2 = cdelt(2)
    call ftpkyf(lun,'CDELT2',cdelt2,8,'Pixel size',status)
    call ftpkys(lun,'CUNIT2','deg','Units of coordinate',status)

    call ftpcom(lun, &
         'This file was produced using the XMM-ESAS software written by', &
         status)
    call ftpcom(lun, &
         'S. L. Snowden, and written using the FITSIO package of ', &
         status)
    call ftpcom(lun,'W. D. Pence.',status)

    !  Define the file format

    call ftpdef(lun,bitpix,naxis,naxes,pcount,gcount,extend, &
         status)
    if((status .ne. 0) .and. (err .eq. 0)) then
       err = 1
       print *, 'STATUS FTPDEF =', status
       print *, 'POSITION 3'
       print *, infile
    endif

    !  Write the file

    if(iflag .eq. 1) then
       call ftp2de(lun,group,isize1,isize1,isize2, &
            rarray,status)
       if((status .ne. 0) .and. (err .eq. 0)) then
          err = 1
          print *, 'STATUS FTP2DE =', status
          print *, 'POSITION 4'
          print *, infile
       endif
    elseif(iflag .eq. 2) then
       call ftp2dj(lun,group,isize1,isize1,isize2, &
            i4array,status)
       if((status .ne. 0) .and. (err .eq. 0)) then
          err = 1
          print *, 'STATUS FTP2DJ =', status
          print *, 'POSITION 6'
          print *, infile
       endif
    endif

    !  Close the file

    call ftclos(lun,status)
    if((status .ne. 0) .and. (err .eq. 0)) then
       err = 1
       print *, 'STATUS FTOPEN =', status
       print *, 'POSITION 7'
       print *, infile
    endif

    !  Clean up

    call ftfiou (lun,status)

    return
  end subroutine putfit

end module gofio_mod
