! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!
module gofmatrix_mod
!
! GOFLIB matrix handling routines
!  
  use dal
  use param 
  use errorHandling

  implicit none
  
contains

  subroutine dcvsrt(covar,ncvm,ma,lista,mfit)

!
!  repacks the covarience matrix for mqmlmn
!
!************************ fform version 1.0 ************  1-mar-88 14:40
!
!  author : press et al.      date:
!  update : j. bloch          date:
!  update : sls               date:  1-mar-1988 14:39
!  update : bhp               date:  2009 rewritten in f95    
!
!  status: tested
!
!   general description 
!  dcvsrt is the real*8 version of routine covsrt found in "numerical
!  recipes" by press et al. see p 515 for details. the only difference
!  is that array covar is real*8 instead of single precision.  given
!  the covarience matrix covar of a fit mfit of ma total parameters,
!  and there ordering lista, repack the covarience matrix to the true
!  order of the parameters.  elements associated with fixed parameters
!  will be zero.  ncvm is the physical dimension of covar
!
!   call_var.          type i/o description 
!  ma                  i4  i   total number of parameters
!  mfit                i4  i   number of active parameters
!  ncvm                i4      dimension of covar array
!  covar               r8  i/o covarience matrix
!  lista               i4  i   position vector
!
!   include_block name          description 
!
!   routines_called    type     description 
!
!   extensions/system calls     description 
! 
!
!***********************************************************************
!
    implicit none

    integer :: i, j
    integer, intent(in) :: ma, mfit, ncvm
    integer, intent(inout), dimension(mfit) :: lista
    real(kind=double), intent(inout), dimension(ncvm,ncvm) :: covar
    real(kind=double) :: swap

    !
    !  zero all elements below the diagonal
    !
    
    do 12 j=1,ma-1
       do 11 i=j+1,ma
          covar(i,j) = 0.d0
11     enddo
12  enddo
    
    !
    !  repack off-diagonal elements of fit into correct locations
    !  below the diagonal
    !
    do 14 i=1,mfit-1
       do 13 j=i+1,mfit
          if(lista(j) .gt. lista(i)) then
             covar(lista(j),lista(i)) = covar(i,j)
          else
             covar(lista(i),lista(j)) = covar(i,j)
          endif
13     enddo
14  enddo
    !
    !  temporarily store origial elements in top row, and zero the diagonal
    !
    swap = covar(1,1)

    do 15 j=1,ma
       covar(1,j) = covar(j,j)
       covar(j,j) = 0.d0
15  enddo

    covar(lista(1),lista(1)) = swap
    !
    !  now sort the elements into proper order on diagonal
    !
    do 16 j=2,mfit
       covar(lista(j),lista(j)) = covar(1,j)
16  enddo
    !
    !  finally, fill in above the diagonal by symmetry
    !
    do 18 j=2,ma
       do 17 i=1,j-1
          covar(i,j) = covar(j,i)
17     enddo
18  enddo

    return

  end subroutine dcvsrt
!
!
!
  subroutine dgausj(a,n,np,b,m,mp,ierr)
!
!  numerical routines linear equation solution routine
!
!************************ fform version 1.0 ************  1-mar-88 15:03
!
!  author : press et al.      date:
!  update : j. bloch          date:  6-sep-1987
!  update : sls               date:  1-mar-1988 14:42
!  update : bhp               date: 2009 rewritten in f95
!  status: tested
!
!  general description 
!  linear equation solution by gauss-jordan elimination.  a is an
!  input matrix of n x n elements, stored in an array of physical
!  dimensions np x np.  b is an input matrix of n x m containing the
!  m right-hand side vectors, stored in an array of physical
!  dimensions np x mp.  on output, a is replaced by its matrix
!  inverse, and b is replaced by the corresponding set of solution
!  vectors.
!  modified by j. bloch to be a double precision matrix version of the
!  subroutine gaussj in the book numerical recipes by press et al.
!
!  call_var.          type i/o description 
!  a                   r8  i/o updated real*8 matrix which will be
!                              replaced by it's inverse.
!  n                   i4  i   actual order of matrix a
!  np                  i4      actual dimenstions of array a.
!  b                   r8  i/o updated array containing m right hand
!                                  side vectors. upon return, b
!                                  contains the solution vectors.
!  m                   i4  i   number of right hand side vectors.
!  mp                  i4  i   actual second dimension size of b.
!  ierr                i4  i/o = 0 no errors
!                              = 3 singular matrix
!
!  include_block name          description 
!
!  routines_called    type     description 
!
!  extensions/system calls     description  
!
!***********************************************************************

    implicit none
    !
    !       parameter nmax=200
    !
    integer :: i, icol, ierr, irow, j, k, l, ll
    integer, intent(in) :: m, mp, n, np
    integer, dimension(200) :: ipiv, indxr, indxc
    real(kind=double), intent(inout) :: b(np,mp), a(np,np)
    real(kind=double) :: big, dum, pivinv

    ierr = 0
    do 11 j=1,n
       ipiv(j) = 0
11  enddo
    !
    !  start the main loop over the columns to be reduced
    !
    i = 0
    do while ((i .lt. n) .and. (ierr .eq. 0))
       i = i + 1
       big = 0.d0
       !
       !  start the outer loop of the search for the pivot element
       !
       j = 0
       do while ((j .lt. n) .and. (ierr .eq. 0))
          j = j + 1
          if(ipiv(j) .ne. 1) then
             k = 0
             do while ((k .lt. n) .and. (ierr .eq. 0))
                k = k + 1
                if (ipiv(k) .eq. 0) then
                   if (abs(a(j,k)) .ge. big) then
                      big = abs(a(j,k))
                      irow = j
                      icol = k
                   endif
                else if (ipiv(k) .gt. 1) then
                   ierr = 3 ! oddly not fatal here, 3=die elsewhere!?
                   call warning("SingularMatrix",&
                        "DGAUSS Singular Matrix. Point 1")
                endif
             enddo
          endif
       enddo
       if(ierr .eq. 0) then
          ipiv(icol) = ipiv(icol) + 1
          !
          !  we now have the pivot element, so we interchange rows, if needed,
          !  to put the pivot element on the diagonal.  the columns are not
          !  physically interchanged, only relabeled: indx(i), the column of the
          !  ith pivot element, is the ith column that is reduced, while indxr(i)
          !  is the row in which that pivot element was origially located.  if
          !  indxr(i) =/ indxc(i) there is an implied column interchange.  with
          !  this form of bookkeeping, the solution b's will end up in the correct
          !  order, and the inverse matrix will be scrambled by columns.
          !
          if (irow .ne. icol) then
             do 14 l=1,n
                dum = a(irow,l)
                a(irow,l) = a(icol,l)
                a(icol,l) = dum
14           enddo

             do 15 l=1,m
                dum = b(irow,l)
                b(irow,l) = b(icol,l)
                b(icol,l) = dum
15           enddo
          endif
          !
          !  we are now ready to divide the pivot row by the pivot element
          !  located at irow and icol
          !
          indxr(i) = irow
          indxc(i) = icol
          if (a(icol,icol) .eq. 0.) then
             ierr = 3 ! also not fatal here??? 
             call warning("SingularMatrix",&
                  "DGAUSS Singular Matrix. Point 2")
          else
             pivinv = 1.d0/a(icol,icol)
             a(icol,icol) = 1.d0
             do 16 l=1,n
                a(icol,l) = a(icol,l)*pivinv
16           enddo

             do 17 l=1,m
                b(icol,l) = b(icol,l)*pivinv
17           enddo
             !
             !  next we reduce the rows, except for the pivot row
             !
             do 21 ll=1,n
                if(ll .ne. icol)then
                   dum = a(ll,icol)
                   a(ll,icol) = 0.
                   do 18 l=1,n
                      a(ll,l) = a(ll,l) - a(icol,l)*dum
18                 enddo
                   do 19 l=1,m
                      b(ll,l) = b(ll,l) - b(icol,l)*dum
19                 enddo
                endif
21           enddo
          endif
       endif
    enddo
    !
    !  this is the end of the main loop over the columns of the reduction.
    !  now unscramble the solution in view of the column interchanges.
    !  we do this by interchanging pairs of columns in the order that the
    !  permutation was built up.
    !
    if(ierr .eq. 0) then
       do 24 l=n,1,-1
          if(indxr(l).ne.indxc(l))then
             do 23 k=1,n
                dum = a(k,indxr(l))
                a(k,indxr(l)) = a(k,indxc(l))
                a(k,indxc(l)) = dum
23           enddo
          endif
24     enddo
    endif
    return
  end subroutine dgausj
  !
  !
  !
  !
  ! BHP: originally in subsn.f90, then in marq_mod.f90 (espfilt only). now here
  !      called ONLY by mos_back, pn_back
  !
  subroutine poly_fit(tempx,tempy,tempu,n,np,ap)

    implicit none
        
    integer :: i, ierr, iflag, iup, iuplim, &
         lista(10), maxitr, n, np, nppo
    integer :: ftype = 1

    real(kind=single) :: a(10), alamda, almlim, ap(10), chisq, deltch, &
         i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, it, &
         ochisq, olamda, t, tempx(n), tempy(n), tempu(n)
    real(kind=single) :: ainit(10) = (/1.0, -0.1, 0.03, -0.01, 0.003, &
         -0.0001, 0.00003, -0.00001, 0.000003, -0.000001/)
    real(kind=double) :: alpha(10,10), covar(10,10)
    character(len=1) :: fix(10) = (/' ', ' ', ' ', ' ', &
         ' ', ' ', ' ',' ', ' ', ' '/)
    character(len=15) :: pshort(10) = (/'I0', 'I1', 'I2', 'I3', &
         'I4', 'I5', 'I6','I7', 'I8', 'I9'/)
    character(len=80) :: pname(10)

    ierr = 0
    lista = 0
    do i=1,10
       a(i) = ainit(i)
    enddo

    it = 0.0
    t = 0.0
    do i=1,n
       if(tempy(i) .ne. 0.0) then
          t = t + tempy(i)
          it = it + 1
       endif
    enddo
    a(1) = t/it
    nppo = np + 1
    do i=0,np ! BHP: old loop set 'fix' and 'pname' but they are never used...
       lista(i+1) = i + 1
    enddo
    if(np+2 .lt. 10) then
       do i=np+2,9
          lista(i+1) = 0
          a(i) = 0.0
       enddo
    endif

    !  *********************************************************************
    !  *                                                                   *
    !  *                       Main Fit Begins Here                        *
    !  *                                                                   *
    !  *********************************************************************

    maxitr = 100
    almlim = 1.E3
    iuplim = 10

    alamda = -1.0
    ochisq = 0.0
    iup = 0
    i = 1
    iflag = 0
    
    do while ((iflag .eq. 0) .and. (i .lt. maxitr))
       i = i + 1
       olamda = alamda
       call mqm (ftype,tempx,tempy,tempu,n,a,10,lista,nppo, &
            covar,alpha,10,chisq,alamda,ierr)
       if(ierr .eq. 3) then
          write(errstr,*)"Bad return from subroutine mqm"
          call message(AppMsg,Always)
          write(errstr,*)"Program stopped in POLY_FIT. Position 01"
          call error("badPoly_fit",trim(errstr))
       endif
       
       deltch = abs(ochisq - chisq)
       
       !  Check for various termination conditions
       
       if((alamda .lt. olamda) .and. (deltch .lt. 0.1)) &
            iflag = 1
       if(iflag .eq. 0) then
          if(alamda .gt. olamda) iup = iup + 1
          if(alamda .lt. olamda) iup = 0
          if(iup .gt. iuplim) iflag = 1
          if(alamda .gt. almlim) iup = iup + 1
       endif
    enddo

    write(*,1010) ' Np = ',np,' N = ',n,' M.L.E. = ',chisq
1010 format(/,1x,a,i2,a,i4,a,g12.4,a,g14.7)

    !  Finish off the fit to get the covarience matrix
 
    alamda = 0.0
    call mqm (ftype,tempx,tempy,tempu,n,a,10,lista,nppo, &
         covar,alpha,10,chisq,alamda,ierr)
    if(ierr .eq. 3) then
       write(errstr,*)"Bad return from subroutine mqm"
       call message(AppMsg,Always)
       write(errstr,*)"Program stopped in POLY_FIT. Position 02"
       call error("badPoly_fit",trim(errstr))
    endif
    !
    !  write out final values and one sigma uncertainties
    !
    write(errstr,*) &
         "POLY_FIT: Final parameter values and the formal standard errors"
    call message(AppMsg,Always)

    do i=1,nppo
       ap(i) = a(i)
       write(errstr,2100)pshort(lista(i)),a(lista(i)), &
            sngl(dsqrt(dabs(covar(lista(i),lista(i)))))
       call message(AppMsg,Always) ! change to Noisy???!!!
2100   format(1x,a15,f13.5,' +/-',f11.5)
    enddo
    
    return
    
  end subroutine poly_fit
  !
  !
  !
  subroutine gfuncs (tempx,aloc,ymodel,dyda,dy2da2,lista,mfit,ma)

!
!  calculates gaussian function distribution and partial derivatives
!
!************************ fform version 1.0 ************ 21-apr-88 08:24
!
!  author : sls               date: 17-august-2005 15:00  
!  update : bhp               date: 2009 rewritten in f95!
!    
!  status: tested
!
!   general description 
!   pfuncs is the model value subroutine for gauss_fit for use with the
!   fitting subroutine mqm.for. 
!
!  call_var.          type i/o description 
!  tempx               r4  i/? x value.
!  aloc                r4  i/? array of model values.
!                              a(1) = normalization
!                              a(2) = width
!                              a(3) = peak point
!  ymodel              r4      value of the function.
!  dyda                r4  i/o array of first partial derivatives
!                              w.r.t. each parameter. (only the
!                              parameters actually being fit are
!                              computed.)
!  dy2da2              r4  i/o array of second partial derivatives
!                              w.r.t. the parameters.  (only those
!                              parameters being fit will be computed.)
!  lista               i4  i   list of parameters being fitted.
!  mfit                i4  i   number of parameters being fit.
!  ma                  i4  i   number of parameters in model.
!                              (should be <= 10)
!
!   include_block name          description          
!
!   routines_called    type     description 
!
!   extensions/system calls     description 
!
    !***********************************************************************
    
    integer :: i, mfit, lista(mfit), ma

    real(kind=single) :: aloc(ma), dy2da2(10,10), dyda(10), tempx, ymodel
    real(kind=double) :: a(10), chan, g, ymod

    do i=1,10
       a(i) = aloc(i)
    enddo
    chan = dble(tempx)
    !
    !  compute the model value
    !
    g   = exp(-a(2)*(chan - a(3))*(chan - a(3)))
    ymod = a(1)*g
    ymodel = sngl(ymod)
    !
    !  compute values on intervals to prepare for computation
    !  of numerical derivatives.
    !
    dyda(1) = g
    dyda(2) = -a(1)*(chan - a(3))*(chan - a(3))*g
    dyda(3) = 2.0*a(1)*a(2)*(chan - a(3))*g
    !
    dy2da2(1,1) = 0.0
    dy2da2(1,2) = -(chan - a(3))*(chan - a(3))*g
    dy2da2(1,3) = 2.0*a(2)*(chan - a(3))*g
    dy2da2(2,1) = -(chan - a(3))*(chan - a(3))*g
    dy2da2(2,2) = a(1)*(chan - a(3))*(chan - a(3))*&
         (chan - a(3))*(chan - a(3))*g
    dy2da2(2,3) = 2.0*a(1)*(chan - a(3))*&
         (1.0 -  a(2)*(chan - a(3))*(chan - a(3)))*g
    dy2da2(3,1) = 2.0*a(2)*(chan - a(3))*g
    dy2da2(3,2) = 2.0*a(1)*(chan - a(3))*&
         (1.0 -  a(2)*(chan - a(3))*(chan - a(3)))*g
    dy2da2(3,3) = 2.0*a(1)*a(2)*&
         (2.0*a(2)*(chan - a(3))*(chan - a(3)) - 1.0)*g
    return
  end subroutine gfuncs
  !
  ! bhp: pfuncs not actually called by any esas routines, 
  !      converted to f95 for consistency of all marquardt
  !      matrix calculations in original subsn code of SLS.
  !
  subroutine pfuncs (tempx,aloc,ymodel,dyda,dy2da2,lista,mfit,ma)
    !
    integer :: m
    integer :: i, ii, mfit, lista(mfit), ma
    real(kind=single) :: aloc(ma), dy2da2(10,10), dyda(10), tempx, ymodel
    real(kind=double) :: a(10), dtime, ymod

    do i=1,10
       a(i) = dble(aloc(i))
    enddo
    dtime = dble(tempx)
    !
    !  compute the model value
    !
    
    ymod = a(1) + a(2)*dtime + a(3)*dtime*dtime + &
         a(4)*dtime*dtime*dtime + &
         a(5)*dtime*dtime*dtime*dtime + &
         a(6)*dtime*dtime*dtime*dtime*dtime + &
         a(7)*dtime*dtime*dtime*dtime*dtime*dtime + &
         a(8)*dtime*dtime*dtime*dtime*dtime*dtime*dtime
    ymodel = sngl(ymod)
    !
    !  compute values on intervals to prepare for computation
    !  of numerical derivatives.
    !
    dyda(1) = 1
    dyda(2) = dtime
    dyda(3) = dtime*dtime
    dyda(4) = dtime*dtime*dtime
    dyda(5) = dtime*dtime*dtime*dtime
    dyda(6) = dtime*dtime*dtime*dtime*dtime
    dyda(7) = dtime*dtime*dtime*dtime*dtime*dtime
    dyda(8) = dtime*dtime*dtime*dtime*dtime*dtime*dtime
    dyda(9) = 0.0
    dyda(10) = 0.0
 
    do i=1,10
       do ii=1,10
          dy2da2(i,ii) = 0.
       enddo
    enddo
    
    return
    
  end subroutine pfuncs
  !
  ! original call: 
  !
  ! SUBROUTINE MQM (FTYPE,TEMPX,TEMPY,TEMPU,NDATA,A,MA, &
  !      LISTA,MFIT,COVAR,ALPHA,NCA,CHISQ,ALAMDA,IERR)
  !
  subroutine mqm (ftype,tempx,tempy,tempu,ndata,a,ma,&
       lista,mfit,covar,alpha,nca,chisq,alamda,ierr)
!
!  maximum likelyhood fitting routine
!
!************************ fform version 1.0 ************  2-mar-88 09:16
!
!  author : j. bloch          date:  4-sep-1987
!  update : j. bloch          date:  6-sep-1987
!  update : sls               date:  2-mar-1988 09:25  
!  update : bhp               date: 2009 rewritten in f95
!
!  status: tested
!
!   general description 
!  mqmlmn was derived from mrqmin in the book "numerical recipes" by
!  press et al. it finds the best parameter fit using the marquardt
!  method applied to the poisson maximum likelihood estimator, as
!  described by cash in the 1979 march 15 astrophysical journal. the
!  sig array has been removed because it is not needed.
!
!   call_var.          type i/o description 
!  ftype               i4      = 1 polynomial function (pfuncs)
!                              = 2 gaussian function (gfuncs)
!  ierr                i4  i/o = 0 no error
!                              = 3
!  tempx               r4  ?/? independent variable data values.
!  tempy               r4  ?/? dependent variable data values.
!  tempu               r4  ?/? dependent variable uncertainty values.
!  ndata               i4      number of data points to fit.
!  a                   r4  i/o model parameter values.
!  ma                  i4  i/? number of model parameter values.
!  lista               i4  i/o list of parameters
!                                  in array a actually being varied.
!  mfit                i4  i/? number of parameters actually being fit.
!  covar               r8  i/o covariance matrix and work
!                                  array used durint the fit.
!  alpha               r8  i/o matrix alpha used in the marquard method
!                                  (see press et al.,pp. 521-528)
!  nca                 i4  i/? actual dimensions of alpha and covar.
!  chisq               r4  i/o output maximum likelihood estimator.
!  alamda              r4  i/o updated real iteration control variable.
!                              must be set to  < 0.0 on first call
!                              of mqmlmn. if a step fails, almada grows
!                              by a factor of 10. on final call, set
!                              almada = 0.0 so curvature matrix is 
!                              returned.
!  pfuncs              r4  ?/? input name of model function, with call
!                                  sequence:
!                                  call pfuncs(x1,a,ymod,dyda,dy2da2,ma)
!                                      with:
!                                      x1 = independent variable at
!                                          which to evaluate the model.
!                                      a(ma) = array of model params.
!                                      ymod = output model value at x1.
!                                      dyda(ma) = vector of first
!                                           derivatives of the model
!                                           w.r.t. the parameters.
!                                      dy2da2(20,20)= matrix of partial
!                                           second derivatives of the
!                                           model w.r.t the parameters.
!                                      lista = list of parameters being 
!                                           fitted.
!                                      mfit  = number of parameters being 
!                                           fit.
!                                      ma = input integer number of
!                                           parameters in the model.
!
!   include_block name          description 
!
!   routines_called    type     description 
!  dcvsrt              sr      repacks the covarience matrix for mqm!
!  dgausj              sr      numerical routines linear equation 
!                              solution routine
!  mrq_cb              sr      computes matrix elements for mqm
!  extensions/system calls     description 
! 
!
!***********************************************************************
!
    implicit none

        integer :: ma
        integer :: ftype, ierr, ierr1, ihit, j, k, kk, lista(ma), &
            mfit, nca, ndata

        real(kind=single) :: a(ma), alamda, atry(10), chisq, ochisq, &
            tempx(ndata), tempy(ndata), tempu(ndata)

        real(kind=double) :: alpha(nca,nca), covar(nca,nca),  &
            beta(10), da(10)
    
        ierr = 0
        
    !
    !  do some initializations
        !
        !print*,'ftype = ',ftype
        !print*,'ndata = ',ndata
        !print*,'a     = ',a
        !print*,'ma    = ',ma
        !print*,'lista = ',lista
        !print*,'mfit  = ',mfit
        !print*,'nca   = ',nca
        !print*,'chisq = ',chisq
        !print*,'almda = ',alamda
        ! turn these back on ???!!!
!    write(errstr,*)'MQM: ndata  = ',ndata
!    call message(AppMsg,Noisy) ! change to Noisy???!!!
!    write(errstr,*)'MQM: a      = ',a
!    call message(AppMsg,Noisy) ! change to Noisy???!!!
!    write(errstr,*)'MQM: chisq  = ',chisq
!    call message(AppMsg,Noisy) ! change to Noisy???!!!
!    write(errstr,*)'MQM: alamda = ',alamda
!    call message(AppMsg,Noisy) ! change to Noisy???!!!

    if(alamda .lt. 0.) then
       kk = mfit + 1
       !
       !  check to see if lista contains a proper permutation of the coeffs.
       !
       j = 0
       do while ((j .lt. ma) .and. (ierr .eq. 0))
          j = j + 1
          ihit = 0
          do 11 k=1,mfit
             if(lista(k) .eq. j) ihit = ihit + 1
11        enddo
          if(ihit .eq. 0) then
             lista(kk) = j
             kk = kk + 1
          else if (ihit .gt. 1) then
             ierr = 3
             write(errstr,*)"MQM: Improper Permutation in LISTA array. Point 1. ihit = ",ihit
             call warning("ImpPerm",trim(adjustl(errstr)))
          endif
       enddo
       !
       if(kk .ne. (ma+1)) then
          ierr = 3
          write(errstr,*)"MQM: Improper Permutation in LISTA array. Point 2. kk,ma+1 = ",kk,ma+1
          call warning("ImpPerm",trim(adjustl(errstr)))
       endif
       
       if(ierr .eq. 0) then
          alamda = 0.001
          call mrq (ftype,tempx,tempy,tempu,ndata,a,ma,&
               lista,mfit,alpha,beta,nca,chisq,ierr1)
          if(ierr1 .eq. 3) then
             ierr = 3
             write(errstr,*) "MQM: Bad return from MRQ. ierr1 = 3"
             call warning("badMRQ",trim(adjustl(errstr)))
          else
             ochisq = chisq
             do 13 j=1,ma
                atry(j) = a(j)
13           enddo
          endif
       endif
    endif
    !
    !  alter linearized fitting matrix by augmenting diagonal elements
    !
    if(ierr .eq. 0) then
       do 15 j=1,mfit
          do 14 k=1,mfit
             covar(j,k) = alpha(j,k)
14        enddo
          covar(j,j) = alpha(j,j)*(1.0d0 + dble(alamda))
          da(j) = beta(j)
15     enddo
       !
       !  get the matrix solution
       !
       call dgausj(covar,mfit,nca,da,1,1,ierr1)
       if(ierr1 .eq. 3) then
          ierr = 3 ! why???!!!
          write(errstr,*) "MQM: Bad return from DGAUSJ. 1."
          call message(AppMsg,Noisy)
       else
          !
          !  if converged evaluate covarience matrix with alamda = 0
          !
          if(alamda .eq. 0.) then
             call dcvsrt(covar,nca,ma,lista,mfit)
          else
             !
             !  check to see if the trial succeeded
             !
             do 16 j=1,mfit
                atry(lista(j)) = a(lista(j)) + sngl(da(j))
16           enddo

             call mrq (ftype,tempx,tempy,tempu,ndata,&
                  atry,ma,lista,mfit,covar,da,nca,chisq,ierr1)
             if(ierr1 .eq. 3) then
                ierr = 3
                write(errstr,*) &
                     "MQM: Bad return from MRQ. 2."
                call warning("badMRQ",trim(adjustl(errstr)))
             else
                if(chisq .lt. ochisq) then
                   !
                   !  success, accept the new solution
                   !
                   alamda = 0.1*alamda
                   ochisq = chisq
                   
                   do 18 j=1,mfit
                      do 17 k=1,mfit
                         alpha(j,k) = covar(j,k)
17                    enddo
                      beta(j) = da(j)
                      a(lista(j)) = atry(lista(j))
18                 enddo
                else
                   !
                   !  failure, increase alamda and return
                   !
                   alamda = 10.*alamda
                   chisq = ochisq
                endif
             endif
          endif
       endif
    endif
    return
  end subroutine mqm
  !
  !
  !
  subroutine mrq (ftype,tempx,tempy,tempu,ndata,a,ma,&
       lista,mfit,alpha,beta,nalp,chisq,ierr)
!
!  computes matrix elements for mqm
!
!************************ fform version 1.0 ************ dd-mmm-yy hh:mm
!
!  author : j. bloch          date:  4-sep-1987
!  update : j. bloch          date:  6-sep-1987
!  update : sls               date:  2-mar-1988 09:23  
!  update : bhp               date: 2009 rewritten in f95
!  status: tested
!
!   general description:
!  mrq was derived from mrqcof in the book "numerical recipes"
!  by press et al. it computes the matrix elements needed for
!  the marquardt method applied to the poisson maximum likelihood
!  estimator, as described by cash in the 1979 march 15
!  astrophysical journal. the sig array has been removed because
!  it is not needed.
!
!   call_var.          type i/o description 
!  ftype               i4      = 1 polynomial function (pfuncs)
!                              = 2 gaussian function (gfuncs)
!  ierr                i4    o = 0 no error
!                              = 3 illegal model value (model < 0.)
!  tempx               r4  ?/? array of independent var. data values.
!  tempy               r4  ?/? array of dependent variable data values.
!  tempu               r4  ?/? array of dependent variable uncertainty values.
!  ndata               i4      number of data points to fit.
!  a                   r4  i/o array of model parameter values.
!  ma                  i4  i/? number of model parameter values.
!  lista               i4  i/o array list of parameters
!                                  in array a actually being varied.
!  mfit                i4  i/? number of parameters actually being fit.
!  covar               r8  i/o covariance matrix and work
!                                  array used durint the fit.
!  alpha               r8  i/o matrix alpha used in the marquard 
!                              method.  (see press et al., pp. 521-528)
!  beta                        vector beta used in the marquardt 
!                              method.  (see press et al. pp. 521-528)
!  nalp                        actual dimensions of alpha and beta.
!  chisq               r4  i/o output maximum likelihood estimator.
!
!   include_block_name          description
!  r$common:cgenl.cmn          general parameter common block
!
!   routines_called    type     description
!  hflag               r       output flag handling routine
!  wrflag              sr      writes to flag output stream
!  pfuncs              sr      input name of model function, with call
!                                  sequence:
!                              call pfuncs(x1,a,ymod,dyda,dy2da2,lista,
!                                          mfit,ma)
!                                  with:
!                                    x1 = independent variable at
!                                        which to evaluate the model.
!                                    a(ma) = array of model parameters.
!                                    ymod = output model value at x1.
!                                    dyda(ma) = vector of first
!                                         derivatives of the model
!                                         w.r.t. the parameters.
!                                    dy2da2(20,20) = matrix of partial
!                                         second derivatives of the
!                                         model w.r.t the parameters.
!                                    lista(mfit) = same as above.
!                                    mfit = same as above.
!                                    ma = input integer number of
!                                         parameters in the model.
!
!   extensions/system calls     description
!  
!
    !***********************************************************************
    
    implicit none

    integer :: mfit
    integer :: ftype, i, ierr, j, k, kk, lista(mfit), &
         ma, nalp, ndata
    !
    real(kind=single) ::  a(ma), chisq, dy, dyda(10), d2yda2(10,10), &
         tempx(ndata), tempy(ndata), sig2i, tempu(ndata), wt, ymod

    real(kind=double) ::  alpha(nalp,nalp), beta(ma) 

    ierr = 0
    !
    !  initialize (symmetric) alpha and beta
    !
    do 12 j=1,mfit
       do 11 k=1,j
          alpha(j,k) = 0.d0
11     enddo
       beta(j) = 0.d0
12  enddo

    chisq = 0.
    !
    !  summation loop over all data
    !
    i = 0
    do while ((i .lt. ndata) .and. (ierr .eq. 0))
       i = i + 1
       if(ftype .eq. 1) then
          call pfuncs(tempx(i),a,ymod,dyda,d2yda2,lista,mfit,ma)
       elseif(ftype .eq. 2) then
          call gfuncs(tempx(i),a,ymod,dyda,d2yda2,lista,mfit,ma)
       else
          call error("illegalFtype","MRQ: Illegal model type.")
       endif
       if(ymod .lt. -1.0d200) then
          write(errstr,*) "MRQ: Illegal model value =", ymod
          call message(AppMsg,Verbose)
          write(errstr,*) "MRQ:           tempx(i) =", tempx(i)
          call message(AppMsg,Verbose)
          write(errstr,*) "MRQ:           tempy(i) =", tempy(i)
          call message(AppMsg,Verbose)
          write(errstr,*) "MRQ:           tempu(i) =", tempu(i)
          call message(AppMsg,Verbose)
          write(errstr,*) "MRQ:                  a =",&
               (a(kk),kk=1,ma)
          call message(AppMsg,Verbose)
          ierr = 3
       else
          sig2i = 1./(tempu(i)*tempu(i))
          dy = tempy(i) - ymod
          
          do 14 j=1,mfit
             wt = sig2i*dyda(lista(j))
             do 13 k=1,j
                alpha(j,k) = alpha(j,k) + wt*dyda(lista(k))
13           enddo
             beta(j) = beta(j) + dy*wt
14        enddo
          chisq = chisq + dy*dy*sig2i
       endif
    enddo
    !
    if(ierr .eq. 0) then
       !
       !  fill in the symmetric side
       !
       do j=2,mfit
          do k=1,j-1
             alpha(k,j) = alpha(j,k)
          enddo
       enddo
    endif

    return
    
  end subroutine mrq
  !
  !
  !
  subroutine amoeba(p,y,mp,np,ndim,ftol,iter)

    integer :: i, j, n, ndim, iter, mpts, ilo, ihi, inhi
    integer, intent(inout) :: mp, np 
    integer, parameter :: nmax=20, itmax=500
    real(kind=double), parameter :: alpha=1.d0, beta=0.5d0, gamma=2.d0
    real(kind=double) :: ftol, rtol, ypr, yprr
    real(kind=double), dimension(:,:) :: p(mp,np)
    real(kind=double), dimension(:) :: y(mp), pt(mp), pr(nmax), &
         prr(nmax), pbar(nmax)
    mpts = ndim + 1
    iter = 0
    ! Expected maximum number of dimensions, three parameters which define
    ! the expansions and contractions, and maximum allowed number of
    ! iterations.
    loop: do
       if(y(1).gt.y(2)) then
          ihi=1
          inhi=2
       else
          ihi=2
          inhi=1
       endif
       do i=1, mpts
          if(y(i).lt.y(ilo)) ilo=i
          if(y(i).gt.y(ihi)) then
             inhi=ihi
             ihi=i
          else if (y(i).gt.y(inhi)) then
             if(i.ne.ihi) inhi=i
          end if
       end do
       !
       ! compute the fractional range from highest to lowest and return if
       ! satisfactory.
       !
       rtol=2.d0*dabs(y(ihi)-y(ilo))/(dabs(y(ihi))+dabs(y(ilo)))
       call message(AppMsg,Always) ! change to Noisy!!!
       if(rtol.lt.ftol) return
       if(iter.eq.itmax) then
          write(errstr,*) "iter ",iter," GT itmax ",itmax,' exceeded max iterations'
          call message(AppMsg,Noisy)
          call error("iterGTitmax","amoeba exceeded max iterations")
       endif
       iter=iter+1
       do j=1, ndim
          pbar(j)=0.d0
       end do
       do i=1, mpts
          if(i.ne.ihi) then
             do j=1,ndim
                pbar(j)=pbar(j) + p(i,j)
             end do
          end if
       end do
       do j=1, ndim
          pbar(j)=pbar(j)/ndim
          pr(j)=(1.d0+alpha)*pbar(j) - alpha*p(ihi,j)
       end do
       !ypr=func(pr)
       if(ypr.le.y(ilo)) then
          do j=1,ndim
             prr(j)=gamma*pr(j) + (1.d0-gamma)*pbar(j)
          end do
          !yprr=func(prr)
          if(yprr.lt.y(ilo)) then
             do j=1, ndim
                p(ihi,j)=prr(j)
             end do
             y(ihi)=yprr
          else
             do j=1, ndim
                p(ihi,j)=pr(j)
             end do
             y(ihi)=ypr	  
          end if
       else if(ypr.ge.y(inhi)) then
          if(ypr.lt.y(ihi)) then
             do j=1, ndim
                p(ihi,j)=pr(j)
             end do
             y(ihi)=ypr
          end if
          do j=1, ndim
             prr(j)=beta*p(ihi,j) + (1.d0-beta)*pbar(j)
          end do
          !yprr=func(prr)
          if(yprr.lt.y(ihi)) then
             do j=1, ndim
                p(ihi,j)=prr(j)
             end do
             y(ihi)=yprr
          else
             do i=1, mpts
                if(i.ne.ilo) then
                   do j=1,ndim
                      pr(j)=0.5d0*(p(i,j) + p(ilo,j))
                      p(i,j)=pr(j)
                   end do
!                   y(i)=func(pr)
                end if
             end do
          end if
       else
          do j=1, ndim
             p(ihi,j)=pr(j)
          end do
          y(ihi)=ypr
       end if
    end do loop
    
  end subroutine amoeba

  subroutine subamoeba (ndim,vertex,valueN,mask,tolr,maxiter,iter,gauss_ind,gauss_dep,width)
    !
    ! ndim = size of parms array, = 3    ! 
    ! vertex(ndim,ndim+1), input matrix vertex(3,4)
    ! valueN(ndim+1), input vector
    ! tolr = fractional convergence tolerance = 0.01
    ! maxiter = max number of iterations performed = 100
    ! iter  = current iteration
    ! gauss_ind = gaussian independent vector
    ! gauss_dep = gaussian dependent vector
    ! width = size of gaussian vectors
    ! 
    ! alpha = reflection parameter
    ! beta  = contraction parameter
    ! gamma = expansion parameter
    !
    real(kind=double), parameter :: alpha = 1.d0, beta = 0.5d0, gamma = 2.d0
    integer, intent(in) :: maxiter, width, ndim
    integer, intent(inout) :: iter
    integer :: numvert, aflag, nsflag, puflag, mask(3)
    integer :: nchans_lc, i, j , k, hi, lo, nhi, nlo, itt
    real(kind=double), intent(inout) :: valueN(ndim+1), tolr
    real(kind=double), allocatable, dimension(:) :: pbar, parms, newvert, &
         newnewvert
    real(kind=double) :: vertex(ndim,ndim+1)
    real(kind=double) :: gauss_ind(width), gauss_dep(width)
    real(kind=double) :: rtol, newval, newnewval, summ    
    !setup
    numvert = ndim+1 ! 4
    iter    = 0
    nsflag  = 0 
    puflag  = ndim ! size(parms)
    aflag   = 0
    allocate(pbar(ndim),parms(ndim),newvert(ndim),newnewvert(ndim+1))
    itt = 0
    start: do
       itt = itt + 1
       lo = 2 !!! check this ???
       do j=1,ndim ! 3
          summ = 0.d0
          do i = 1,numvert ! 4
             summ = summ + vertex(j,i)
          enddo
          parms(j) = summ / 4.d0
       enddo
       if (valueN(1) .gt. valueN(2)) then
          hi = 1
          nhi = 2
       else
          hi = 2
          nhi = 1
       end if
       !
       ! find point with the highest chisqr
       !
       do i=1,numvert
          if (valueN(i) .lt. valueN(lo)) lo = i
          if (valueN(i) .gt. valueN(hi)) then
             nhi = hi
             hi = i
          elseif (valueN(i) .gt. valueN(nhi)) then
             if (i .ne. hi) nhi = i
          end if
       end do
       !
       ! compute fractional range and test for return
       !
       if (iter >= maxiter) then
          iter = -1
          return
       end if
       rtol = 2.0*dabs(valueN(hi)-valueN(lo))/(dabs(valueN(hi))+dabs(valueN(lo)))
       if((rtol .lt. tolr) .and. (iter .gt. 0)) then 
          if(nsflag .ne. 0) then
             return
          endif
       endif
       puflag=ndim ! 3
       do i=1,ndim ! 0,mdim = 0,2 in IDL
          if(mask(i) .ne. 0) then 
             if(2.0*dabs(vertex(i,hi)-pbar(i))/(dabs(vertex(i,hi))+dabs(pbar(i))) .lt. 1.d-8) then 
                puflag = puflag - 1
             endif
          endif
       end do
       if(puflag .lt. 2) then
          iter=-iter
          return
       endif
       nsflag = 0
       iter = iter + 1
       pbar = 0.d0
       do i=1,numvert
          if(i.ne.hi) then
             do j=1,ndim
                pbar(j)=pbar(j) + vertex(j,i)
             end do
          end if
       end do
       pbar = pbar / real(ndim,8) 
       do i=1,ndim
          newvert(i)= real(mask(i),8) * (1.d0+alpha)*pbar(i) - alpha*vertex(i,hi) !tamoeba
       end do
       call fit_gauss(width,newvert,gauss_ind,gauss_dep,newval) ! return with newval
       if(newval.le.valueN(lo)) then
          do j=1,ndim
             newnewvert(j)=gamma*newvert(j) + (1.d0-gamma)*pbar(j)
          end do
          call fit_gauss(width,newnewvert,gauss_ind,gauss_dep,newnewval) ! return with newnewval
          if(newnewval.lt.valueN(lo)) then
             do j=1, ndim
                vertex(j,hi)=newnewvert(j)
             end do
             valueN(hi)=newnewval
          else
             do j=1, ndim
                vertex(j,hi)=newvert(j)
             end do
             valueN(hi)=newval	  
          end if
       else if(newval.ge.valueN(nhi)) then
          if(newval.lt.valueN(hi)) then
             do j=1, ndim
                vertex(j,hi)=newvert(j)
             end do
             valueN(hi)=newval
             aflag=1
          end if
          do j=1, ndim
             newnewvert(j)=beta*vertex(j,hi) + (1.d0-beta)*pbar(j)
          end do
          call fit_gauss(width,newnewvert,gauss_ind,gauss_dep,newnewval) ! return with newnewval
          if(newnewval.lt.valueN(hi)) then
             do j=1, ndim
                vertex(j,hi)=newnewvert(j)
             end do
             valueN(hi)=newnewval
             aflag=0
          else
             do i=1, numvert
                if(i.ne.lo) then
                   nsflag=1
                   do j=1,ndim
                      newvert(j)=0.5d0*(vertex(j,i) + vertex(j,lo))
                      vertex(j,i)=newvert(j)
                   end do
                   call fit_gauss(width,newnewvert,gauss_ind,gauss_dep,valueN(i)) ! return with valueN(i)
                end if
                aflag=0
             end do
          end if
          aflag=0
       else
          do j=1, ndim
             vertex(j,hi)=newvert(j)
          end do
          valueN(hi)=newval
       end if

    end do start

    return

  end subroutine subamoeba
  !
  !
  !
  subroutine kpamoeba(ndim,parms,dparms,mask,tolr,maxiter,reslo,reshi,iter,&
       chilo,chihi,gauss_ind,gauss_dep,width)
    !
    ! rewrite of kipamoeba.pro IDL routine, June, 2020 by BP
    !inputs
    !  parms=initial parameters for fitting functions
    !  dparms=distance along each axis in phase space to be searched
    !  mask=mask for disabbling change to a parameter
    !        changeable parameters=1, unchangeable parameters=0
    !  tolr=(max(dep-fit)-min(dep-fit))/avg(dep-fit)
    !  maxiter=max number of iterations to execute
    !outputs
    !  reslo=set of best parameters
    !  reshi=set of worst allowable parameters
    !  iter=number of iterations run (-1 if maxed out without hitting tolr)
    !  chilo=chisqr for best fit
    !  chihi=chisqr for worst allowable fit
    !  width=high-low+1 from clean_lc
    !
    integer :: iter, ndim, i, j, k, maxiter, subiter, nn, summ
    integer :: rank(1), bbb, itt
    integer, intent(in) :: mask(3), width

    real(kind=double) :: fact, chisqr, rtol, tolr, chihi, chilo, oldchisqr
    real(kind=double) :: vertex(ndim,ndim+1)
    real(kind=double) :: valueN(ndim+1), res(ndim), temp(ndim)
    real(kind=double), intent(inout) :: parms(ndim), dparms(ndim), &
         reslo(ndim), reshi(ndim)
    real(kind=double) :: gauss_ind(width), gauss_dep(width)
    !
    iter = 0
    fact = -1.D0
    vertex = 0.d0
    call fit_gauss(width,parms,gauss_ind,gauss_dep,chisqr)
    itt = 0
1   loop: do
       itt = itt + 1
       do i=1,ndim
          vertex(i,1) = parms(i) ! IDL: vertex(0,0)=parms
       end do
       call fit_gauss(width,parms,gauss_ind,gauss_dep,valueN(1))
       fact = fact*(-1.d0)
       temp = 0.d0
       !
       do i=ndim-1,ndim+1  ! check indices !!! IDL = 1,3, so 2,4 for f90
          do j=1,ndim
             vertex(j,i) = parms(j)
          end do
          vertex(i-1,i) = vertex(i-1,i) + dparms(i-1) * real(mask(i-1),8) * fact !!! if 1,1,1 ???
          do j=1,ndim
             temp(j) = vertex(j,i) ! temp=vertex(*,i)
          end do
          call fit_gauss(width,temp,gauss_ind,gauss_dep,valueN(i))
       end do
       call subamoeba(ndim,vertex,valueN,mask,tolr,maxiter,subiter,gauss_ind,gauss_dep,width)

       if (subiter .eq. -1) subiter = maxiter
       iter = iter + subiter
       oldchisqr = chisqr
       rank = minloc(valueN) !!! fix this???

       do i=1,ndim
          res(i) = vertex(i,rank(1)) ! works! 
       end do

       chisqr = valueN(rank(1))
       rtol=dabs(2.0d0*(oldchisqr-chisqr)/(oldchisqr+chisqr))
       if (rtol .gt. tolr) then
          do i=1,ndim
             parms(i) = res(i)
          end do
          go to 1
       else
          exit
       endif

    end do loop

    do i=1,ndim
       reshi(i) = vertex(i,rank(1))
       reslo(i) = res(i)
    end do
    call fit_gauss(width,reshi,gauss_ind,gauss_dep,chihi)
    chilo = chisqr

    return

  end subroutine kpamoeba
  !
  !
  subroutine fit_gauss(nn,parms,gauss_ind,gauss_dep,chi)

    ! rewrite of fit_gauss.pro IDL routine, June, 2020 by BP

    integer :: nn, i, j, k
    real(kind=double), intent(inout) :: gauss_ind(nn), gauss_dep(nn)
    real(kind=double) :: res(nn), parms(3)
    real(kind=double), intent(out) :: chi
    !
    ! e.g. parms(hist_cnts(maxp), hist_rate(maxp), 0.1) 
    ! gauss_ind = hist_rate(low:high) from clean_lc
    ! gauss_dep = hist_cnts(low:high) from clean_lc
    !
    chi = 0.d0
    parms = abs(parms) ! why would they ever be negative???
    res=parms(1) * exp (-(gauss_ind-parms(2))**2/2./parms(3)/parms(3))
    chi=sum((gauss_dep-res)**2)/size(res)
    return    
  end subroutine fit_gauss

end module gofmatrix_mod
